System.register(['roslib', 'three', 'three/examples/jsm/loaders/STLLoader.js', 'three/examples/jsm/loaders/ColladaLoader.js', 'three-mesh-bvh', 'three-spritetext', '@juggle/resize-observer', 'lodash.debounce', 'three-freeform-controls', 'randomcolor'], function (exports, module) {
  'use strict';
  var ROSLIB, Group$1, Quaternion, Color, Mesh$1, Math$1, BoxGeometry, MeshStandardMaterial, MeshPhongMaterial, CylinderGeometry, ConeGeometry, Line$1, Geometry, Vector3, LineBasicMaterial, VertexColors$1, TorusGeometry, FontLoader, TextGeometry, MeshBasicMaterial, three, DefaultLoadingManager, SphereGeometry, PointsMaterial, BufferGeometry, BufferAttribute, Points$2, LineSegments$1, FaceColors, DoubleSide, Face3, Object3D, TextureLoader, PlaneGeometry, FrontSide, CanvasTexture, NearestFilter, Scene$1, DirectionalLight, AmbientLight, GridHelper, Texture, ShaderMaterial, EventDispatcher, Box3, Matrix3, Vector2, Spherical, Sphere$1, OrthographicCamera, WebGLRenderer, PerspectiveCamera, Euler, STLLoader, ColladaLoader, threeMeshBvh, SpriteText, ResizeObserver$2, debounce, ANCHOR_MODE, DEFAULT_HANDLE_GROUP_NAME, ControlsManager, RAYCASTER_EVENTS, randomColor;
  return {
    setters: [function (module) {
      ROSLIB = module.default;
    }, function (module) {
      Group$1 = module.Group;
      Quaternion = module.Quaternion;
      Color = module.Color;
      Mesh$1 = module.Mesh;
      Math$1 = module.Math;
      BoxGeometry = module.BoxGeometry;
      MeshStandardMaterial = module.MeshStandardMaterial;
      MeshPhongMaterial = module.MeshPhongMaterial;
      CylinderGeometry = module.CylinderGeometry;
      ConeGeometry = module.ConeGeometry;
      Line$1 = module.Line;
      Geometry = module.Geometry;
      Vector3 = module.Vector3;
      LineBasicMaterial = module.LineBasicMaterial;
      VertexColors$1 = module.VertexColors;
      TorusGeometry = module.TorusGeometry;
      FontLoader = module.FontLoader;
      TextGeometry = module.TextGeometry;
      MeshBasicMaterial = module.MeshBasicMaterial;
      three = module.default;
      DefaultLoadingManager = module.DefaultLoadingManager;
      SphereGeometry = module.SphereGeometry;
      PointsMaterial = module.PointsMaterial;
      BufferGeometry = module.BufferGeometry;
      BufferAttribute = module.BufferAttribute;
      Points$2 = module.Points;
      LineSegments$1 = module.LineSegments;
      FaceColors = module.FaceColors;
      DoubleSide = module.DoubleSide;
      Face3 = module.Face3;
      Object3D = module.Object3D;
      TextureLoader = module.TextureLoader;
      PlaneGeometry = module.PlaneGeometry;
      FrontSide = module.FrontSide;
      CanvasTexture = module.CanvasTexture;
      NearestFilter = module.NearestFilter;
      Scene$1 = module.Scene;
      DirectionalLight = module.DirectionalLight;
      AmbientLight = module.AmbientLight;
      GridHelper = module.GridHelper;
      Texture = module.Texture;
      ShaderMaterial = module.ShaderMaterial;
      EventDispatcher = module.EventDispatcher;
      Box3 = module.Box3;
      Matrix3 = module.Matrix3;
      Vector2 = module.Vector2;
      Spherical = module.Spherical;
      Sphere$1 = module.Sphere;
      OrthographicCamera = module.OrthographicCamera;
      WebGLRenderer = module.WebGLRenderer;
      PerspectiveCamera = module.PerspectiveCamera;
      Euler = module.Euler;
    }, function (module) {
      STLLoader = module.default;
    }, function (module) {
      ColladaLoader = module.default;
    }, function (module) {
      threeMeshBvh = module.default;
    }, function (module) {
      SpriteText = module.default;
    }, function (module) {
      ResizeObserver$2 = module.ResizeObserver;
    }, function (module) {
      debounce = module.default;
    }, function (module) {
      ANCHOR_MODE = module.ANCHOR_MODE;
      DEFAULT_HANDLE_GROUP_NAME = module.DEFAULT_HANDLE_GROUP_NAME;
      ControlsManager = module.ControlsManager;
      RAYCASTER_EVENTS = module.RAYCASTER_EVENTS;
    }, function (module) {
      randomColor = module.default;
    }],
    execute: function () {

      class Core {
        constructor(ros, topicName, messageType, options = {}) {
          const { onHeaderChange } = options;
          this.options = options;
          this.ros = ros;
          this.headerFrameId = '';
          if (ros && topicName) {
            this.changeTopic(topicName, messageType, false);
          }
          this.onHeaderChange = onHeaderChange || (() => {});
          this.update = this.update.bind(this);
        }

        hide() {
          this.object.visible = false;
        }

        show() {
          this.object.visible = true;
        }

        destroy() {
          this.unsubscribe();
          if (this.object && this.object.parent) {
            this.object.parent.remove(this.object);
            this.object = null;
          }
        }

        reset() {}

        subscribe() {
          if (!this.topicInstances) {
            return;
          }
          this.topicInstances.forEach(t => {
            t.subscribe(this.update);
          });
        }

        unsubscribe() {
          if (!this.topicInstances) {
            return;
          }
          this.topicInstances.forEach(t => {
            t.unsubscribe();
          });
        }

        update(message) {
          const header = message.header ? message.header.frame_id : '';
          if (header !== this.headerFrameId) {
            this.headerFrameId = header;
            this.onHeaderChange(this.headerFrameId);
          }
        }

        updateOptions(options) {
          this.options = {
            ...this.options,
            ...options,
          };
        }

        changeTopic(topicName, type, autoSubscribe = true) {
          const { compression, queueSize, throttleRate } = this.options;

          if (autoSubscribe) {
            this.unsubscribe();
          }

          this.topicName = topicName;
          this.messageType = type || this.messageType;
          this.topicInstances = (Array.isArray(topicName)
            ? topicName
            : [{ name: topicName, messageType: type }]
          ).map(
            ({ name, messageType }) =>
              new ROSLIB.Topic({
                ros: this.ros,
                name,
                messageType,
                compression: compression || 'none',
                throttle_rate: throttleRate || 0,
                queue_size: queueSize || 10,
              }),
          );

          if (autoSubscribe) {
            this.reset();
            this.subscribe();
          }
        }
      }

      /* eslint-disable prettier/prettier */

      const OBJECT_TYPE_ARROW = 'Arrow';
      const OBJECT_TYPE_ARROW_WITH_CIRCLE = 'ArrowWithCircle';
      const OBJECT_TYPE_AXES = 'Axes';
      const OBJECT_TYPE_FLAT_ARROW = 'FlatArrow';

      const MAX_POINTCLOUD_POINTS = 5000000;

      const DEFAULT_BACKGROUND_COLOR = '#000000';
      const DEFAULT_GRID_SIZE = 30;
      const DEFAULT_GRID_DIVISIONS = 30;
      const DEFAULT_GRID_COLOR = '#222222';
      const DEFAULT_GRID_COLOR_CENTERLINE = '#333333';

      const INTERACTIVE_MARKER_ORIENTATION_MODES = {
        'INHERIT': 0,
        'FIXED': 1,
        'VIEW_FACING': 2,
      };

      const UNSUPPORTED_INTERACTIVE_MARKER_ORIENTATION_MODES = [
        INTERACTIVE_MARKER_ORIENTATION_MODES.VIEW_FACING,
      ];

      const INTERACTIVE_MARKER_INTERACTION_MODES = {
        'NONE': 0,
        'MENU': 1,
        'BUTTON': 2,
        'MOVE_AXIS': 3,
        'MOVE_PLANE': 4,
        'ROTATE_AXIS': 5,
        'MOVE_ROTATE': 6,
        'MOVE_3D': 7,
        'ROTATE_3D': 8,
        'MOVE_ROTATE_3D': 9,
      };

      /* * ***************************
       *   Message types
       * ************************** */
      const MESSAGE_TYPE_ROBOT_MODEL = 'robot_description';
      const MESSAGE_TYPE_ACCELSTAMPED = 'geometry_msgs/AccelStamped';
      const MESSAGE_TYPE_POINTSTAMPED = 'geometry_msgs/PointStamped';
      const MESSAGE_TYPE_POLYGONSTAMPED = 'geometry_msgs/PolygonStamped';
      const MESSAGE_TYPE_POSEARRAY = 'geometry_msgs/PoseArray';
      const MESSAGE_TYPE_POSECOVARIANCE = 'geometry_msgs/PoseWithCovariance';
      const MESSAGE_TYPE_POSESTAMPED = 'geometry_msgs/PoseStamped';
      const MESSAGE_TYPE_TWISTSTAMPED = 'geometry_msgs/TwistStamped';
      const MESSAGE_TYPE_VECTOR3STAMPED = 'geometry_msgs/Vector3Stamped';
      const MESSAGE_TYPE_WRENCHSTAMPED = 'geometry_msgs/WrenchStamped';

      const MESSAGE_TYPE_BOUNDINGVOLUME = 'moveit_msgs/BoundingVolume';
      const MESSAGE_TYPE_COLLISION_OBJECT = 'moveit_msgs/CollisionObject';
      const MESSAGE_TYPE_DISPLAYROBOTSTATE = 'moveit_msgs/DisplayRobotState';
      const MESSAGE_TYPE_DISPLAYTRAJECTORY = 'moveit_msgs/DisplayTrajectory';
      const MESSAGE_TYPE_PLANNINGSCENE = 'moveit_msgs/PlanningScene';

      const MESSAGE_TYPE_OCCUPANCYGRID = 'nav_msgs/OccupancyGrid';
      const MESSAGE_TYPE_ODOMETRY = 'nav_msgs/Odometry';
      const MESSAGE_TYPE_PATH = 'nav_msgs/Path';

      const MESSAGE_TYPE_COMPRESSEDIMAGE = 'sensor_msgs/CompressedImage';
      const MESSAGE_TYPE_DISPLAYJOINTSTATE = 'sensor_msgs/JointState';
      const MESSAGE_TYPE_IMAGE = 'sensor_msgs/Image';
      const MESSAGE_TYPE_LASERSCAN = 'sensor_msgs/LaserScan';
      const MESSAGE_TYPE_MAGNETICFIELD = 'sensor_msgs/MagneticField';
      const MESSAGE_TYPE_POINTCLOUD = 'sensor_msgs/PointCloud';
      const MESSAGE_TYPE_POINTCLOUD2 = 'sensor_msgs/PointCloud2';
      const MESSAGE_TYPE_RANGE = 'sensor_msgs/Range';

      const MESSAGE_TYPE_TF = 'tf/tfMessage';
      const MESSAGE_TYPE_TF2 = 'tf2_msgs/TFMessage';

      const MESSAGE_TYPE_MARKER = 'visualization_msgs/Marker';
      const MESSAGE_TYPE_MARKERARRAY = 'visualization_msgs/MarkerArray';
      const MESSAGE_TYPE_INTERACTIVEMARKER = 'visualization_msgs/InteractiveMarkerInit';
      const MESSAGE_TYPE_INTERACTIVEMARKER_UPDATE = 'visualization_msgs/InteractiveMarkerUpdate';
      const MESSAGE_TYPE_INTERACTIVEMARKER_FEEDBACK = 'visualization_msgs/InteractiveMarkerFeedback';

      /** ***************************
       *   Visualizations
       * ************************** */
      const VIZ_TYPE_IMAGE = 'Image';
      const VIZ_TYPE_INTERACTIVEMARKER = 'InteractiveMarker';
      const VIZ_TYPE_LASERSCAN = 'LaserScan';
      const VIZ_TYPE_MAP = 'Map';
      const VIZ_TYPE_MARKER = 'Marker';
      const VIZ_TYPE_MARKERARRAY = 'MarkerArray';
      const VIZ_TYPE_ODOMETRY = 'Odometry';
      const VIZ_TYPE_PATH = 'Path';
      const VIZ_TYPE_POINT = 'Point';
      const VIZ_TYPE_POINTCLOUD = 'PointCloud';
      const VIZ_TYPE_POLYGON = 'Polygon';
      const VIZ_TYPE_POSE = 'Pose';
      const VIZ_TYPE_POSEARRAY = 'PoseArray';
      const VIZ_TYPE_RANGE = 'Range';
      const VIZ_TYPE_ROBOTMODEL = 'RobotModel';
      const VIZ_TYPE_TF = 'Tf';
      const VIZ_TYPE_WRENCH = 'Wrench';

      /* * ***************************
       *   Viz specific constants
       * ************************** */

      const POINT_FIELD_DATATYPES = {
        INT8: 1,
        UINT8: 2,
        INT16: 3,
        UINT16: 4,
        INT32: 5,
        UINT32: 6,
        FLOAT32: 7,
        FLOAT64: 8,
      };

      const COLLISION_OBJECT_OPERATIONS = {
        ADD: 0,
        REMOVE: 1,
        APPEND: 2,
        MOVE: 3,
      };

      const SOLID_PRIMITIVE_TYPES = {
        BOX: 1,
        SPHERE: 2,
        CYLINDER: 3,
        CONE: 4,
      };

      const MARKER_OBJECT_TYPES = {
        ARROW: 0,
        CUBE: 1,
        SPHERE: 2,
        CYLINDER: 3,
        LINE_STRIP: 4,
        LINE_LIST: 5,
        CUBE_LIST: 6,
        SPHERE_LIST: 7,
        POINTS: 8,
        TEXT_VIEW_FACING: 9,
        MESH_RESOURCE: 10,
        TRIANGLE_LIST: 11,
      };

      const LINE_STYLES = {
        LINES: 'Lines',
        BILLBOARDS: 'Billboards',
      };

      const MAP_COLOR_SCHEMES = {
        MAP: 'map',
        CONST_MAP: 'constmap',
        RAW: 'raw',
      };

      const LASERSCAN_STYLES = {
        SQUARES: 'squares',
        POINTS: 'points',
        FLAT_SQUARES: 'flat_squares',
        SPHERES: 'spheres',
        BOXES: 'boxes',
      };

      const COLOR_TRANSFORMERS = {
        INTENSITY: 'Intensity',
        AXIS_COLOR: 'AxisColor',
        FLAT_COLOR: 'FlatColor',
      };

      const DEPTHCLOUD_STREAMTYPES = {
        VP8: 'vp8',
        MJPEG: 'mjpeg',
      };

      const POINTCLOUD_COLOR_CHANNELS = {
        RGB: 'rgb',
        INTENSITY: 'intensity',
      };

      const AXES = {
        X: 'x',
        Y: 'y',
        Z: 'z',
      };

      const INTENSITY_CHANNEL_OPTIONS = {
        INTENSITY: 'intensity',
        ...AXES,
      };

      const ODOMETRY_OBJECT_TYPES = {
        arrow: OBJECT_TYPE_ARROW,
        axes: OBJECT_TYPE_AXES,
      };

      const POSE_OBJECT_TYPES = {
        arrow: OBJECT_TYPE_ARROW,
        axes: OBJECT_TYPE_AXES,
        flatArrow: OBJECT_TYPE_FLAT_ARROW,
      };

      const WRENCH_OBJECT_TYPES = {
        arrow: OBJECT_TYPE_ARROW,
        arrowWithCircle: OBJECT_TYPE_ARROW_WITH_CIRCLE,
      };





      /** ***************************
       *   Default Options
       * ************************** */
      const DEFAULT_CYLINDER_HEIGHT = 1;
      const DEFAULT_CYLINDER_RADIUS = 1;
      const DEFAULT_RADIAL_SEGMENTS = 32;

      const DEFAULT_CONE_HEIGHT = 1;
      const DEFAULT_CONE_RADIUS = 1;

      const DEFAULT_COLOR_X_AXIS = '#ff0000';
      const DEFAULT_COLOR_Y_AXIS = '#008000';
      const DEFAULT_COLOR_Z_AXIS = '#0000ff';
      const DEFAULT_COLOR_ARROW = '#f0ff00';
      const DEFAULT_COLOR_LINE = '#f0ff00';

      const DEFAULT_OPTIONS_SCENE = {
        backgroundColor: DEFAULT_BACKGROUND_COLOR,
        gridSize: DEFAULT_GRID_SIZE,
        gridDivisions: DEFAULT_GRID_DIVISIONS,
        gridColor: DEFAULT_GRID_COLOR,
        gridCenterlineColor: DEFAULT_GRID_COLOR_CENTERLINE,
      };

      const DEFAULT_OPTIONS_TF_VIEWER = {
        selectedFrame: '',
      };

      const DEFAULT_OPTIONS_ARROW = {
        shaftLength: 1,
        shaftRadius: 0.05,
        headLength: 0.3,
        headRadius: 0.1,
      };

      const DEFAULT_OPTIONS_AXES = {
        axesLength: 1,
        axesRadius: 0.1,
      };

      const DEFAULT_OPTIONS_FLATARROW = {
        arrowLength: 0.3,
      };

      const DEFAULT_OPTIONS_DISPLAYTRAJECTORY = {
        robot: new Group$1(),
        loop: true,
      };

      const DEFAULT_OPTIONS_INTENSITY = {
        channelName: INTENSITY_CHANNEL_OPTIONS.INTENSITY,
        useRainbow: false,
        invertRainbow: false,
        minColor: '#000000',
        maxColor: '#ffffff',
        autocomputeIntensityBounds: false,
        maxIntensity: 3730,
        minIntensity: 388,
      };

      const DEFAULT_OPTIONS_COLLISION_OBJECT = {};

      const DEFAULT_OPITONS_AXIS_COLOR = {
        axis: AXES.X,
        autocomputeValueBounds: false,
        useFixedFrame: false,
        minAxisValue: 0,
        maxAxisValue: 0,
      };

      const DEFAULT_OPTIONS_DEPTHCLOUD = {
        streamType: DEPTHCLOUD_STREAMTYPES.MJPEG,
        f: 526,
        maxDepthPerTile: 1.0,
        pointSize: 3,
        width: 1024,
        height: 1024,
        whiteness: 0,
        varianceThreshold: 0.000016667,
      };

      const DEFAULT_OPTIONS_IMAGE = {
        queueSize: 1,
        width: 320,
        height: 240,
        compression: 'cbor',
      };

      const DEFAULT_OPTIONS_IMAGE_STREAM = {
        width: 320,
        height: 240,
      };

      const DEFAULT_OPTIONS_LASERSCAN = {
        selectable: false,
        style: LASERSCAN_STYLES.FLAT_SQUARES,
        size: 0.05,
        alpha: 1,
        decayTime: 0,
        queueSize: 10,
        compression: 'cbor',
        colorTransformer: COLOR_TRANSFORMERS.INTENSITY,
        flatColor: '#ffffff',
        ...DEFAULT_OPTIONS_INTENSITY,
        ...DEFAULT_OPITONS_AXIS_COLOR,
      };

      const DEFAULT_OPTIONS_MAP = {
        alpha: 1,
        colorScheme: MAP_COLOR_SCHEMES.MAP,
        compression: 'cbor',
        drawBehind: false,
      };

      const DEFAULT_OPTIONS_MARKER = {
        queueSize: 1,
        namespaces: [],
      };

      const DEFAULT_OPTIONS_MARKERARRAY = {
        queueSize: 1,
        namespaces: [],
        throttleRate: 0,
      };

      const DEFAULT_OPTIONS_INTERACTIVE_MARKER = {
        queueSize: 1,
        namespaces: [],
        throttleRate: 0,
        publishThrottleRate: 0,
        hideOtherHandlesOnSelect: true,
        hideOtherControlsInstancesOnSelect: true,
      };

      const DEFAULT_OPTIONS_ODOMETRY = {
        type: OBJECT_TYPE_ARROW,
        color: DEFAULT_COLOR_X_AXIS,
        alpha: 1,
        ...DEFAULT_OPTIONS_ARROW,
        ...DEFAULT_OPTIONS_AXES,
        ...DEFAULT_OPTIONS_FLATARROW,
        positionTolerance: 0.1,
        angleTolerance: 0.1,
        keep: 100,
      };

      const DEFAULT_OPTIONS_PATH = {
        color: '#ffffff',
        alpha: 1,
      };

      const DEFAULT_OPTIONS_PLANNINGSCENE = {};

      const DEFAULT_OPTIONS_POINTCLOUD = {
        compression: 'cbor',
        colorChannel: POINTCLOUD_COLOR_CHANNELS.RGB,
        size: 0.0125,
        useRainbow: true,
        queueSize: 1,
      };

      const DEFAULT_OPTIONS_POINT = {
        color: '#ff0000',
        alpha: 1,
        radius: 0.2,
      };

      const DEFAULT_OPTIONS_POLYGON = {
        color: '#ffffff',
        alpha: 1,
      };

      const DEFAULT_OPTIONS_POSE = {
        color: DEFAULT_COLOR_X_AXIS,
        alpha: 1,
        ...DEFAULT_OPTIONS_ARROW,
        ...DEFAULT_OPTIONS_AXES,
        type: POSE_OBJECT_TYPES.arrow,
      };

      const DEFAULT_OPTIONS_TORUS = {
        circleRadius: 0.3,
        tube: 0.03,
        radialSegments: 16,
        tubularSegments: 100,
        arc: 1.7*Math.PI,
      };

      const DEFAULT_OPTIONS_ARROW_WITH_CIRCLE = {
        ...DEFAULT_OPTIONS_ARROW,
        ...DEFAULT_OPTIONS_TORUS,
        circleConeLength: 0.1,
        circleConeRadius: 0.1,
      };

      const DEFAULT_OPTIONS_WRENCH = {
          forceColor: DEFAULT_COLOR_X_AXIS,
          torqueColor: DEFAULT_COLOR_ARROW,
          alpha: 1,
          forceArrowScale: 1,
          torqueArrowScale: 1,
          arrowWidth: 1,
          ...DEFAULT_OPTIONS_ARROW_WITH_CIRCLE,
          type: WRENCH_OBJECT_TYPES.arrow,
      };

      const DEFAULT_OPTIONS_POSEARRAY = {
        color: DEFAULT_COLOR_X_AXIS,
        alpha: 1,
        ...DEFAULT_OPTIONS_ARROW,
        ...DEFAULT_OPTIONS_AXES,
        ...DEFAULT_OPTIONS_FLATARROW,
        type: POSE_OBJECT_TYPES.arrow,
      };

      const DEFAULT_OPTIONS_RANGE = {
        color: DEFAULT_COLOR_X_AXIS,
        alpha: 1,
      };

      const DEFAULT_OPTIONS_ROBOTMODEL = {};

      const DEFAULT_OPTIONS_TF = {};

      /** ***************************
       *   Zethus exports
       * ************************** */
      const SUPPORTED_MESSAGE_TYPES = [
        MESSAGE_TYPE_IMAGE,
        MESSAGE_TYPE_LASERSCAN,
        MESSAGE_TYPE_MARKER,
        MESSAGE_TYPE_MARKERARRAY,
        MESSAGE_TYPE_OCCUPANCYGRID,
        MESSAGE_TYPE_ODOMETRY,
        MESSAGE_TYPE_PATH,
        MESSAGE_TYPE_PLANNINGSCENE,
        MESSAGE_TYPE_POINTCLOUD2,
        MESSAGE_TYPE_POLYGONSTAMPED,
        MESSAGE_TYPE_POSESTAMPED,
        MESSAGE_TYPE_POSEARRAY,
        MESSAGE_TYPE_POSESTAMPED,
        MESSAGE_TYPE_ROBOT_MODEL,
        MESSAGE_TYPE_TF,
        MESSAGE_TYPE_TF2,
      ];

      const UNSUPPORTED_MESSAGE_TYPES = [
        MESSAGE_TYPE_ACCELSTAMPED,
        MESSAGE_TYPE_BOUNDINGVOLUME,
        MESSAGE_TYPE_COMPRESSEDIMAGE,
        MESSAGE_TYPE_DISPLAYJOINTSTATE,
        MESSAGE_TYPE_DISPLAYROBOTSTATE,
        MESSAGE_TYPE_DISPLAYTRAJECTORY,
        MESSAGE_TYPE_INTERACTIVEMARKER,
        MESSAGE_TYPE_MAGNETICFIELD,
        MESSAGE_TYPE_POINTCLOUD,
        MESSAGE_TYPE_POSECOVARIANCE,
        MESSAGE_TYPE_RANGE,
        MESSAGE_TYPE_TWISTSTAMPED,
        MESSAGE_TYPE_VECTOR3STAMPED,
        MESSAGE_TYPE_WRENCHSTAMPED,
      ];

      var CONSTANTS = /*#__PURE__*/Object.freeze({
        __proto__: null,
        OBJECT_TYPE_ARROW: OBJECT_TYPE_ARROW,
        OBJECT_TYPE_ARROW_WITH_CIRCLE: OBJECT_TYPE_ARROW_WITH_CIRCLE,
        OBJECT_TYPE_AXES: OBJECT_TYPE_AXES,
        OBJECT_TYPE_FLAT_ARROW: OBJECT_TYPE_FLAT_ARROW,
        MAX_POINTCLOUD_POINTS: MAX_POINTCLOUD_POINTS,
        DEFAULT_BACKGROUND_COLOR: DEFAULT_BACKGROUND_COLOR,
        DEFAULT_GRID_SIZE: DEFAULT_GRID_SIZE,
        DEFAULT_GRID_DIVISIONS: DEFAULT_GRID_DIVISIONS,
        DEFAULT_GRID_COLOR: DEFAULT_GRID_COLOR,
        DEFAULT_GRID_COLOR_CENTERLINE: DEFAULT_GRID_COLOR_CENTERLINE,
        INTERACTIVE_MARKER_ORIENTATION_MODES: INTERACTIVE_MARKER_ORIENTATION_MODES,
        UNSUPPORTED_INTERACTIVE_MARKER_ORIENTATION_MODES: UNSUPPORTED_INTERACTIVE_MARKER_ORIENTATION_MODES,
        INTERACTIVE_MARKER_INTERACTION_MODES: INTERACTIVE_MARKER_INTERACTION_MODES,
        MESSAGE_TYPE_ROBOT_MODEL: MESSAGE_TYPE_ROBOT_MODEL,
        MESSAGE_TYPE_ACCELSTAMPED: MESSAGE_TYPE_ACCELSTAMPED,
        MESSAGE_TYPE_POINTSTAMPED: MESSAGE_TYPE_POINTSTAMPED,
        MESSAGE_TYPE_POLYGONSTAMPED: MESSAGE_TYPE_POLYGONSTAMPED,
        MESSAGE_TYPE_POSEARRAY: MESSAGE_TYPE_POSEARRAY,
        MESSAGE_TYPE_POSECOVARIANCE: MESSAGE_TYPE_POSECOVARIANCE,
        MESSAGE_TYPE_POSESTAMPED: MESSAGE_TYPE_POSESTAMPED,
        MESSAGE_TYPE_TWISTSTAMPED: MESSAGE_TYPE_TWISTSTAMPED,
        MESSAGE_TYPE_VECTOR3STAMPED: MESSAGE_TYPE_VECTOR3STAMPED,
        MESSAGE_TYPE_WRENCHSTAMPED: MESSAGE_TYPE_WRENCHSTAMPED,
        MESSAGE_TYPE_BOUNDINGVOLUME: MESSAGE_TYPE_BOUNDINGVOLUME,
        MESSAGE_TYPE_COLLISION_OBJECT: MESSAGE_TYPE_COLLISION_OBJECT,
        MESSAGE_TYPE_DISPLAYROBOTSTATE: MESSAGE_TYPE_DISPLAYROBOTSTATE,
        MESSAGE_TYPE_DISPLAYTRAJECTORY: MESSAGE_TYPE_DISPLAYTRAJECTORY,
        MESSAGE_TYPE_PLANNINGSCENE: MESSAGE_TYPE_PLANNINGSCENE,
        MESSAGE_TYPE_OCCUPANCYGRID: MESSAGE_TYPE_OCCUPANCYGRID,
        MESSAGE_TYPE_ODOMETRY: MESSAGE_TYPE_ODOMETRY,
        MESSAGE_TYPE_PATH: MESSAGE_TYPE_PATH,
        MESSAGE_TYPE_COMPRESSEDIMAGE: MESSAGE_TYPE_COMPRESSEDIMAGE,
        MESSAGE_TYPE_DISPLAYJOINTSTATE: MESSAGE_TYPE_DISPLAYJOINTSTATE,
        MESSAGE_TYPE_IMAGE: MESSAGE_TYPE_IMAGE,
        MESSAGE_TYPE_LASERSCAN: MESSAGE_TYPE_LASERSCAN,
        MESSAGE_TYPE_MAGNETICFIELD: MESSAGE_TYPE_MAGNETICFIELD,
        MESSAGE_TYPE_POINTCLOUD: MESSAGE_TYPE_POINTCLOUD,
        MESSAGE_TYPE_POINTCLOUD2: MESSAGE_TYPE_POINTCLOUD2,
        MESSAGE_TYPE_RANGE: MESSAGE_TYPE_RANGE,
        MESSAGE_TYPE_TF: MESSAGE_TYPE_TF,
        MESSAGE_TYPE_TF2: MESSAGE_TYPE_TF2,
        MESSAGE_TYPE_MARKER: MESSAGE_TYPE_MARKER,
        MESSAGE_TYPE_MARKERARRAY: MESSAGE_TYPE_MARKERARRAY,
        MESSAGE_TYPE_INTERACTIVEMARKER: MESSAGE_TYPE_INTERACTIVEMARKER,
        MESSAGE_TYPE_INTERACTIVEMARKER_UPDATE: MESSAGE_TYPE_INTERACTIVEMARKER_UPDATE,
        MESSAGE_TYPE_INTERACTIVEMARKER_FEEDBACK: MESSAGE_TYPE_INTERACTIVEMARKER_FEEDBACK,
        VIZ_TYPE_IMAGE: VIZ_TYPE_IMAGE,
        VIZ_TYPE_INTERACTIVEMARKER: VIZ_TYPE_INTERACTIVEMARKER,
        VIZ_TYPE_LASERSCAN: VIZ_TYPE_LASERSCAN,
        VIZ_TYPE_MAP: VIZ_TYPE_MAP,
        VIZ_TYPE_MARKER: VIZ_TYPE_MARKER,
        VIZ_TYPE_MARKERARRAY: VIZ_TYPE_MARKERARRAY,
        VIZ_TYPE_ODOMETRY: VIZ_TYPE_ODOMETRY,
        VIZ_TYPE_PATH: VIZ_TYPE_PATH,
        VIZ_TYPE_POINT: VIZ_TYPE_POINT,
        VIZ_TYPE_POINTCLOUD: VIZ_TYPE_POINTCLOUD,
        VIZ_TYPE_POLYGON: VIZ_TYPE_POLYGON,
        VIZ_TYPE_POSE: VIZ_TYPE_POSE,
        VIZ_TYPE_POSEARRAY: VIZ_TYPE_POSEARRAY,
        VIZ_TYPE_RANGE: VIZ_TYPE_RANGE,
        VIZ_TYPE_ROBOTMODEL: VIZ_TYPE_ROBOTMODEL,
        VIZ_TYPE_TF: VIZ_TYPE_TF,
        VIZ_TYPE_WRENCH: VIZ_TYPE_WRENCH,
        POINT_FIELD_DATATYPES: POINT_FIELD_DATATYPES,
        COLLISION_OBJECT_OPERATIONS: COLLISION_OBJECT_OPERATIONS,
        SOLID_PRIMITIVE_TYPES: SOLID_PRIMITIVE_TYPES,
        MARKER_OBJECT_TYPES: MARKER_OBJECT_TYPES,
        LINE_STYLES: LINE_STYLES,
        MAP_COLOR_SCHEMES: MAP_COLOR_SCHEMES,
        LASERSCAN_STYLES: LASERSCAN_STYLES,
        COLOR_TRANSFORMERS: COLOR_TRANSFORMERS,
        POINTCLOUD_COLOR_CHANNELS: POINTCLOUD_COLOR_CHANNELS,
        AXES: AXES,
        INTENSITY_CHANNEL_OPTIONS: INTENSITY_CHANNEL_OPTIONS,
        ODOMETRY_OBJECT_TYPES: ODOMETRY_OBJECT_TYPES,
        POSE_OBJECT_TYPES: POSE_OBJECT_TYPES,
        WRENCH_OBJECT_TYPES: WRENCH_OBJECT_TYPES,
        DEFAULT_CYLINDER_HEIGHT: DEFAULT_CYLINDER_HEIGHT,
        DEFAULT_CYLINDER_RADIUS: DEFAULT_CYLINDER_RADIUS,
        DEFAULT_RADIAL_SEGMENTS: DEFAULT_RADIAL_SEGMENTS,
        DEFAULT_CONE_HEIGHT: DEFAULT_CONE_HEIGHT,
        DEFAULT_CONE_RADIUS: DEFAULT_CONE_RADIUS,
        DEFAULT_COLOR_X_AXIS: DEFAULT_COLOR_X_AXIS,
        DEFAULT_COLOR_Y_AXIS: DEFAULT_COLOR_Y_AXIS,
        DEFAULT_COLOR_Z_AXIS: DEFAULT_COLOR_Z_AXIS,
        DEFAULT_COLOR_ARROW: DEFAULT_COLOR_ARROW,
        DEFAULT_COLOR_LINE: DEFAULT_COLOR_LINE,
        DEFAULT_OPTIONS_SCENE: DEFAULT_OPTIONS_SCENE,
        DEFAULT_OPTIONS_TF_VIEWER: DEFAULT_OPTIONS_TF_VIEWER,
        DEFAULT_OPTIONS_ARROW: DEFAULT_OPTIONS_ARROW,
        DEFAULT_OPTIONS_AXES: DEFAULT_OPTIONS_AXES,
        DEFAULT_OPTIONS_FLATARROW: DEFAULT_OPTIONS_FLATARROW,
        DEFAULT_OPTIONS_DISPLAYTRAJECTORY: DEFAULT_OPTIONS_DISPLAYTRAJECTORY,
        DEFAULT_OPTIONS_COLLISION_OBJECT: DEFAULT_OPTIONS_COLLISION_OBJECT,
        DEFAULT_OPTIONS_DEPTHCLOUD: DEFAULT_OPTIONS_DEPTHCLOUD,
        DEFAULT_OPTIONS_IMAGE: DEFAULT_OPTIONS_IMAGE,
        DEFAULT_OPTIONS_IMAGE_STREAM: DEFAULT_OPTIONS_IMAGE_STREAM,
        DEFAULT_OPTIONS_LASERSCAN: DEFAULT_OPTIONS_LASERSCAN,
        DEFAULT_OPTIONS_MAP: DEFAULT_OPTIONS_MAP,
        DEFAULT_OPTIONS_MARKER: DEFAULT_OPTIONS_MARKER,
        DEFAULT_OPTIONS_MARKERARRAY: DEFAULT_OPTIONS_MARKERARRAY,
        DEFAULT_OPTIONS_INTERACTIVE_MARKER: DEFAULT_OPTIONS_INTERACTIVE_MARKER,
        DEFAULT_OPTIONS_ODOMETRY: DEFAULT_OPTIONS_ODOMETRY,
        DEFAULT_OPTIONS_PATH: DEFAULT_OPTIONS_PATH,
        DEFAULT_OPTIONS_PLANNINGSCENE: DEFAULT_OPTIONS_PLANNINGSCENE,
        DEFAULT_OPTIONS_POINTCLOUD: DEFAULT_OPTIONS_POINTCLOUD,
        DEFAULT_OPTIONS_POINT: DEFAULT_OPTIONS_POINT,
        DEFAULT_OPTIONS_POLYGON: DEFAULT_OPTIONS_POLYGON,
        DEFAULT_OPTIONS_POSE: DEFAULT_OPTIONS_POSE,
        DEFAULT_OPTIONS_TORUS: DEFAULT_OPTIONS_TORUS,
        DEFAULT_OPTIONS_ARROW_WITH_CIRCLE: DEFAULT_OPTIONS_ARROW_WITH_CIRCLE,
        DEFAULT_OPTIONS_WRENCH: DEFAULT_OPTIONS_WRENCH,
        DEFAULT_OPTIONS_POSEARRAY: DEFAULT_OPTIONS_POSEARRAY,
        DEFAULT_OPTIONS_RANGE: DEFAULT_OPTIONS_RANGE,
        DEFAULT_OPTIONS_ROBOTMODEL: DEFAULT_OPTIONS_ROBOTMODEL,
        DEFAULT_OPTIONS_TF: DEFAULT_OPTIONS_TF,
        SUPPORTED_MESSAGE_TYPES: SUPPORTED_MESSAGE_TYPES,
        UNSUPPORTED_MESSAGE_TYPES: UNSUPPORTED_MESSAGE_TYPES
      });

      const setTransform = (
        object,
        {
          translation: { x: posX, y: posY, z: posZ },
          rotation: { w: orientW, x: orientX, y: orientY, z: orientZ },
        },
      ) => {
        object.position.set(posX, posY, posZ);
        object.quaternion.copy(
          new Quaternion(orientX, orientY, orientZ, orientW).normalize(),
        );
      };

      const setScale = (object, { x, y, z }) => {
        object.scale.set(x, y, z);
      };

      const setColor = (object, color) => {
        if (typeof color === 'string') {
          object.material.color = new Color(color);
        } else {
          object.material.color.setRGB(color.r, color.g, color.b);
        }
      };

      class Group extends Group$1 {
        setTransform(transform) {
          setTransform(this, transform);
        }

        setScale(scale) {
          setScale(this, scale);
        }

        setColor(colors) {
          this.children.forEach(child => {
            setColor(child, colors);
          });
        }
      }

      class Mesh extends Mesh$1 {
        setTransform(transform) {
          setTransform(this, transform);
        }

        setScale(scale) {
          setScale(this, scale);
        }

        setColor(colors) {
          setColor(this, colors);
        }

        setAlpha(alpha) {
          this.material.opacity = Math$1.clamp(alpha, 0, 1);
        }
      }

      class Box extends Mesh {
        constructor() {
          super();
          this.geometry = new BoxGeometry();
          this.material = new MeshStandardMaterial();
        }
      }

      class CollisionObject extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_COLLISION_OBJECT) {
          super(ros, topicName, MESSAGE_TYPE_COLLISION_OBJECT, {
            ...DEFAULT_OPTIONS_COLLISION_OBJECT,
            ...options,
          });

          this.object = new Group();
          this.updateOptions({
            ...DEFAULT_OPTIONS_COLLISION_OBJECT,
            ...options,
          });
        }

        update(message) {
          super.update(message);
          const { id, operation, primitive_poses: poses, primitives } = message;
          const existingObject = this.object.getObjectByName(id);
          switch (operation) {
            case COLLISION_OBJECT_OPERATIONS.ADD: {
              if (existingObject) {
                existingObject.parent.remove(existingObject);
              }
              const newObject = new Group();
              newObject.name = id;
              primitives.forEach((primitiveInfo, index) => {
                const primitive = CollisionObject.getNewPrimitive(primitiveInfo);
                primitive.setTransform({
                  translation: poses[index].position,
                  rotation: poses[index].orientation,
                });
                newObject.add(primitive);
              });
              this.object.add(newObject);
              break;
            }
            case COLLISION_OBJECT_OPERATIONS.REMOVE: {
              if (existingObject) {
                existingObject.parent.remove(existingObject);
              }
              break;
            }
            case COLLISION_OBJECT_OPERATIONS.APPEND: {
              primitives.forEach(primitiveInfo => {
                const primitive = CollisionObject.getNewPrimitive(primitiveInfo);
                primitive.setTransform({
                  translation: primitiveInfo.position,
                  rotation: primitiveInfo.orientation,
                });
                existingObject.add(primitive);
              });
              break;
            }
          }
        }

        static getNewPrimitive({ type, dimensions }) {
          switch (type) {
            case SOLID_PRIMITIVE_TYPES.BOX:
              const primitive = new Box();
              const [x, y, z] = dimensions;
              primitive.setScale({ x, y, z });
              return primitive;
            default:
              return null;
          }
        }
      }

      class DisplayTrajectory extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_DISPLAYTRAJECTORY) {
          super(ros, topicName, MESSAGE_TYPE_DISPLAYTRAJECTORY, options);

          this.object = new Group();
          this.updateOptions({
            ...DEFAULT_OPTIONS_DISPLAYTRAJECTORY,
            ...options,
          });
          const { robot } = this.options;
          this.robotClone = robot.clone(true);
          Object.keys(this.robotClone.links).forEach(linkName => {
            const link = this.robotClone.links[linkName];
            link.traverse(child => {
              if (child.material) {
                child.material = new MeshPhongMaterial({ color: '#ff0000' });
              }
            });
          });
          this.lastMessage = null;
          this.loopbackId = null;
          this.poseRemovalId = null;
          this.pointsUpdateIds = [];
        }

        updateOptions(options) {
          super.updateOptions(options);
          if (options.loop && this.lastMessage) {
            this.update(this.lastMessage, true);
          } else {
            this.resetLoopback();
          }
        }

        resetLoopback() {
          clearTimeout(this.loopbackId);
          clearTimeout(this.poseRemovalId);
          if (this.pointsUpdateIds) {
            this.pointsUpdateIds.map(x => clearTimeout(x));
          }
          this.pointsUpdateIds = [];
          if (this.robotClone && this.robotClone.parent) {
            this.robotClone.parent.remove(this.robotClone);
          }
        }

        update(message, loopback) {
          const { loop } = this.options;
          this.resetLoopback();
          if (!loopback) {
            this.lastMessage = message;
          }

          super.update(message);
          const {
            trajectory: [
              {
                joint_trajectory: { joint_names: jointNames, points },
              },
            ],
            trajectory_start: {
              joint_state: { name: initialNames, position: initialPositions },
            },
          } = message;
          this.object.add(this.robotClone);
          initialNames.forEach((name, index) => {
            const joint = this.robotClone.getObjectByName(name);
            if (joint) {
              joint.setAngle(initialPositions[index]);
            }
          });
          points.forEach(point => {
            const {
              positions,
              time_from_start: { nsecs, secs },
            } = point;
            this.pointsUpdateIds.push(
              setTimeout(() => {
                jointNames.forEach((jointName, index) => {
                  const joint = this.robotClone.getObjectByName(jointName);
                  if (joint) {
                    joint.setAngle(positions[index]);
                  }
                });
              }, 1000 * secs + nsecs / 1000000),
            );
          });
          if (points.length > 0) {
            const {
              time_from_start: { nsecs: lastNsec, secs: lastSec },
            } = points[points.length - 1];
            this.poseRemovalId = setTimeout(() => {
              this.robotClone.parent.remove(this.robotClone);
              if (loop) {
                this.loopbackId = setTimeout(() => {
                  this.update(this.lastMessage, true);
                }, 1000);
              }
            }, 1000 * lastSec + lastNsec / 1000000);
          }
        }
      }

      class PlanningScene extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_PLANNINGSCENE) {
          super(ros, topicName, MESSAGE_TYPE_PLANNINGSCENE, {
            ...DEFAULT_OPTIONS_PLANNINGSCENE,
            ...options,
          });

          this.object = new Group();
          this.collisionObjectViz = new CollisionObject();
          this.attachedCollisionObjects = new Map();
          this.object.add(this.collisionObjectViz.object);
          this.updateOptions({
            ...MESSAGE_TYPE_PLANNINGSCENE,
            ...options,
          });
        }

        update(message) {
          super.update(message);
          const {
            robot_state: {
              attached_collision_objects: attachedCollisionObjects = [],
              is_diff: isRobotStateDiff,
              joint_state: { name, position },
            },
            world: { collision_objects: worldCollisionObjects },
          } = message;
          worldCollisionObjects.forEach(collisionMessage => {
            this.collisionObjectViz.update(collisionMessage);
          });
          if (attachedCollisionObjects.length) {
            attachedCollisionObjects.map(({ link_name: linkName, object }) => {
              const collisionVizInstance = new CollisionObject();
              collisionVizInstance.object = this.object.getObjectByName(linkName);
              collisionVizInstance.update(object);
              this.attachedCollisionObjects.set(object.id, collisionVizInstance);
            });
          }
          if (!isRobotStateDiff) {
            const newCollisionObjectNames = attachedCollisionObjects.map(
              ({ object: { id } }) => id,
            );
            for (const [
              collObjName,
              collObj,
            ] of this.attachedCollisionObjects.entries()) {
              if (newCollisionObjectNames.indexOf(collObjName) === -1) {
                collObj.update({
                  id: collObjName,
                  operation: COLLISION_OBJECT_OPERATIONS.REMOVE,
                });
                this.attachedCollisionObjects.delete(collObjName);
              }
            }
          }
          name.forEach((jointName, index) => {
            const joint = this.object.getObjectByName(jointName);
            if (joint) {
              joint.setAngle([position[index]]);
            }
          });
        }
      }

      class Cylinder extends Mesh {
        constructor(
          color = DEFAULT_COLOR_ARROW,
          radius = DEFAULT_CYLINDER_RADIUS,
          height = DEFAULT_CYLINDER_HEIGHT,
        ) {
          super();
          this.geometry = new CylinderGeometry(
            radius,
            radius,
            height,
            DEFAULT_RADIAL_SEGMENTS,
          );
          this.material = new MeshStandardMaterial({ color });
          this.material.transparent = true;
          this.rotateX(Math.PI / 2);
        }
      }

      class Cone extends Mesh {
        constructor(color = DEFAULT_COLOR_X_AXIS) {
          super();
          this.geometry = new ConeGeometry(
            DEFAULT_CONE_RADIUS,
            DEFAULT_CONE_HEIGHT,
            DEFAULT_RADIAL_SEGMENTS,
          );
          this.material = new MeshStandardMaterial({ color });
          this.material.transparent = true;
        }
      }

      class Arrow extends Group {
        constructor() {
          super();

          this.cone = new Cone(DEFAULT_COLOR_X_AXIS);
          this.cone.rotateZ(-Math.PI / 2);

          this.cylinder = new Cylinder(DEFAULT_COLOR_X_AXIS);
          this.cylinder.rotateZ(-Math.PI / 2);

          this.cone.setScale({
            x: DEFAULT_CONE_RADIUS,
            y: DEFAULT_CONE_HEIGHT,
            z: DEFAULT_CONE_RADIUS,
          });

          this.cylinder.setScale({
            x: DEFAULT_CYLINDER_RADIUS,
            y: DEFAULT_CYLINDER_HEIGHT,
            z: DEFAULT_CYLINDER_RADIUS,
          });
          this.cylinder.translateY(this.cylinder.scale.y / 2);
          this.cone.translateY(this.cylinder.scale.y + this.cone.scale.y / 2);

          this.type = OBJECT_TYPE_ARROW;
          this.add(this.cone);
          this.add(this.cylinder);
        }

        setColor({ cone, cylinder }) {
          if (cone) {
            this.cone.setColor(cone);
          }

          if (cylinder) {
            this.cylinder.setColor(cylinder);
          }
        }

        setHeadDimensions({ radius, length }) {
          const parsedRadius = parseFloat(radius);
          const parsedLength = parseFloat(length);

          if (parsedRadius) {
            const { y } = this.cone.scale;
            this.cone.setScale({ x: radius, y, z: radius });
          }

          if (parsedLength) {
            const { x, z } = this.cone.scale;
            this.cone.setScale({ x, y: parsedLength, z });
            this.cone.position.set(0, 0, 0);
            this.cone.translateY(this.cylinder.scale.y + parsedLength / 2);
          }
        }

        setShaftDimensions({ radius, length }) {
          const parsedRadius = parseFloat(radius);
          const parsedLength = parseFloat(length);

          if (radius) {
            const { y } = this.cylinder.scale;
            this.cylinder.setScale({ x: parsedRadius, y, z: parsedRadius });
          }

          if (length) {
            const { x, z } = this.cylinder.scale;
            this.cylinder.setScale({ x, y: parsedLength, z });
            this.cylinder.position.set(0, 0, 0);
            this.cylinder.translateY(parsedLength / 2);
            this.setHeadDimensions({ length: this.cone.scale.y });
          }
        }

        setAlpha(alpha) {
          this.cylinder.setAlpha(alpha);
          this.cone.setAlpha(alpha);
        }

        setScale(scale) {
          const { x } = scale;
          const [y, z] = [x / 2, x / 2];
          setScale(this, { x, y, z });
        }
      }

      class Axes extends Group {
        constructor(radius, height) {
          super();
          this.x = new Cylinder(DEFAULT_COLOR_X_AXIS, radius, height);
          this.y = new Cylinder(DEFAULT_COLOR_Y_AXIS, radius, height);
          this.z = new Cylinder(DEFAULT_COLOR_Z_AXIS, radius, height);

          this.x.translateX((height || DEFAULT_CYLINDER_HEIGHT) / 2);
          this.y.translateZ(-(height || DEFAULT_CYLINDER_HEIGHT) / 2);
          this.z.translateY((height || DEFAULT_CYLINDER_HEIGHT) / 2);

          this.x.rotateZ(-Math.PI / 2);
          this.y.rotateX(Math.PI / 2);

          this.type = OBJECT_TYPE_AXES;
          this.add(this.x);
          this.add(this.y);
          this.add(this.z);
        }

        setLength(length) {
          length = parseFloat(length);
          [this.x, this.y, this.z].forEach(axis => {
            axis.position.set(0, 0, 0);
            axis.scale.setY(length);
          });

          this.x.translateY(length / 2);
          this.y.translateY(-length / 2);
          this.z.translateY(length / 2);
        }

        setRadius(radius) {
          this.children.forEach(child => {
            child.scale.setX(parseFloat(radius));
            child.scale.setZ(parseFloat(radius));
          });
        }
      }

      class Line extends Line$1 {
        constructor(color, disableVertexColor) {
          super();
          this.geometry = new Geometry();
          this.geometry.vertices.push(new Vector3(0, 0, 0));
          const colorOptions = {};

          if (!disableVertexColor) {
            colorOptions.vertexColors = VertexColors$1;
          }

          this.material = new LineBasicMaterial({ ...colorOptions });
          this.material.transparent = true;
        }

        setColor(colors) {
          setColor(this, colors);
        }

        updatePoints(points, colors = []) {
          this.geometry.vertices = points.map(({ x, y, z }) => new Vector3(x, y, z));
          this.geometry.verticesNeedUpdate = true;

          const color = [];
          colors.forEach(({ r, g, b }) => {
            color.push(new Color(r, g, b));
          });

          this.geometry.colors = color;
          this.geometry.colorsNeedUpdate = true;
        }

        setTransform(transform) {
          setTransform(this, transform);
        }

        setAlpha(alpha) {
          this.material.opacity = alpha;
        }
      }

      class LineArrow extends Group {
        constructor() {
          super();
          this.type = OBJECT_TYPE_FLAT_ARROW;
          this.arrowTop = new Line(0xff0000, true);
          this.topPoints = [];
          this.topPoints.push(new Vector3(2, 1, 0));
          this.topPoints.push(new Vector3(3, 0, 0));
          this.topPoints.push(new Vector3(2, -1, 0));
          this.arrowTop.updatePoints(this.topPoints);
          this.add(this.arrowTop);

          this.arrowLength = new Line(0xff0000, true);
          this.arrowLength.updatePoints([new Vector3(0, 0, 0), new Vector3(3, 0, 0)]);
          this.add(this.arrowLength);

          this.scale.set(0.1, 0.1, 0.1);
        }

        setLength(length) {
          this.scale.set(length, length, length);
        }

        setColor(color) {
          const { b, g, r } = color;
          this.arrowTop.material.color.setRGB(r, g, b);
          this.arrowLength.material.color.setRGB(r, g, b);
        }
      }

      const checkToleranceThresholdExceed = (oldPose, newPose, options) => {
        const { angleTolerance, positionTolerance } = options;
        const { position, quaternion } = newPose;
        const { position: oldPosition, quaternion: oldQuaternion } = oldPose;

        const positionToleranceBool =
          oldPosition.distanceTo(position) > positionTolerance;
        const angleToleranceBool = oldQuaternion.angleTo(quaternion) > angleTolerance;

        return positionToleranceBool || angleToleranceBool;
      };

      const setObjectDimension = (object, options) => {
        switch (options.type) {
          case OBJECT_TYPE_ARROW: {
            const {
              alpha,
              color,
              headLength,
              headRadius,
              shaftLength,
              shaftRadius,
            } = options;

            object.setHeadDimensions({ radius: headRadius, length: headLength });
            object.setShaftDimensions({ radius: shaftRadius, length: shaftLength });
            object.setAlpha(alpha);
            object.setColor({
              cone: new Color(color),
              cylinder: new Color(color),
            });
            break;
          }
          case OBJECT_TYPE_AXES: {
            const { axesLength, axesRadius } = options;

            object.setLength(axesLength);
            object.setRadius(axesRadius);
            break;
          }
          case OBJECT_TYPE_FLAT_ARROW: {
            const { arrowLength, color } = options;

            object.setLength(arrowLength);
            object.setColor(new Color(color));
            break;
          }
          case OBJECT_TYPE_ARROW_WITH_CIRCLE:
            const {
              alpha,
              circleConeLength,
              circleConeRadius,
              circleRadius,
              color,
              headLength,
              headRadius,
              shaftLength,
              shaftRadius,
              tube,
            } = options;

            object.setHeadDimensions({ radius: headRadius, length: headLength });
            object.setShaftDimensions({ radius: shaftRadius, length: shaftLength });
            object.setTorusDimensions({ radius: circleRadius, tube });
            object.setCircleConeDimensions({
              radius: circleConeRadius,
              length: circleConeLength,
            });
            object.setAlpha({
              cone: alpha,
              cylinder: alpha,
              torus: alpha,
              circleCone: alpha,
            });
            object.setColor({
              cone: new Color(color),
              cylinder: new Color(color),
              torus: new Color(color),
              circleCone: new Color(color),
            });
            break;
        }
      };

      class Pose extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_POSE) {
          super(ros, topicName, MESSAGE_TYPE_POSESTAMPED, {
            ...DEFAULT_OPTIONS_POSE,
            ...options,
          });
          this.object = new Group$1();
          this.primitive = null;
          this.updateOptions({
            ...DEFAULT_OPTIONS_POSE,
            ...options,
          });
        }

        static getNewPrimitive(options) {
          const { type } = options;
          let newObject = null;

          switch (type) {
            case POSE_OBJECT_TYPES.arrow:
              newObject = new Arrow();
              break;
            case POSE_OBJECT_TYPES.axes:
              newObject = new Axes();
              break;
            case POSE_OBJECT_TYPES.flatArrow:
              newObject = new LineArrow();
              break;
          }

          return newObject;
        }

        updateOptions(options) {
          super.updateOptions(options);
          const { type } = this.options;

          if (this.primitive && this.primitive.type !== type) {
            this.object.remove(this.primitive);
            this.primitive = null;
          }

          if (!this.primitive) {
            this.primitive = Pose.getNewPrimitive(this.options);
            this.object.add(this.primitive);
          }

          setObjectDimension(this.primitive, this.options);
        }

        update(message) {
          super.update(message);
          const {
            pose: { orientation, position },
          } = message;
          this.primitive.setTransform({
            translation: position,
            rotation: orientation,
          });
        }
      }

      class ArrowWithCircle extends Arrow {
        constructor() {
          super();

          this.circleCone = new Cone(DEFAULT_COLOR_X_AXIS);
          this.add(this.circleCone);

          const torusOptions = DEFAULT_OPTIONS_TORUS;

          this.material = new MeshStandardMaterial({
            color: this.cone.material.color,
          });
          this.material.transparent = true;

          const torusGeometry = new TorusGeometry(
            torusOptions.circleRadius,
            torusOptions.tube,
            torusOptions.radialSegments,
            torusOptions.tubularSegments,
            torusOptions.arc,
          );

          this.torus = new Mesh$1(torusGeometry, this.material);
          this.add(this.torus);

          this.circleCone.rotateX(this.torus.geometry.parameters.arc);
          this.circleCone.translateZ(
            -Math$1.cos(-this.torus.geometry.parameters.arc) *
              this.torus.geometry.parameters.radius,
          );
          this.circleCone.translateY(
            -Math$1.sin(-this.torus.geometry.parameters.arc) *
              this.torus.geometry.parameters.radius,
          );
          this.circleCone.translateX(this.cylinder.scale.y / 2);

          this.torus.translateX(this.cylinder.scale.y / 2);
          this.torus.rotateY(Math$1.PI / 2);
        }

        setTorusDimensions({ radius, tube }) {
          const torusOptions = DEFAULT_OPTIONS_TORUS;

          this.torus.geometry = new TorusGeometry(
            radius,
            tube,
            torusOptions.radialSegments,
            torusOptions.tubularSegments,
            torusOptions.arc,
          );

          this.torus.position.set(0, 0, 0);
          this.torus.rotation.set(0, 0, 0);
          this.torus.translateX(this.cylinder.scale.y / 2);
          this.torus.rotateY(Math$1.PI / 2);
        }

        setCircleConeDimensions({ radius, length }) {
          const parsedRadius = parseFloat(radius);
          const parsedLength = parseFloat(length);

          if (parsedRadius) {
            const { y } = this.circleCone.scale;
            this.circleCone.setScale({ x: radius, y, z: radius });
          }

          if (parsedLength) {
            const { x, z } = this.circleCone.scale;
            this.circleCone.setScale({ x, y: parsedLength, z });
          }

          this.circleCone.position.set(0, 0, 0);
          this.circleCone.rotation.set(0, 0, 0);
          this.circleCone.translateX(this.cylinder.scale.y / 2);
          this.circleCone.translateZ(
            -Math$1.cos(-this.torus.geometry.parameters.arc) *
              this.torus.geometry.parameters.radius,
          );
          this.circleCone.translateY(
            -Math$1.sin(-this.torus.geometry.parameters.arc) *
              this.torus.geometry.parameters.radius,
          );
          this.circleCone.rotateX(this.torus.geometry.parameters.arc);
        }

        setColor({ cone, cylinder, torus, circleCone }) {
          if (cone) {
            this.cone.setColor(cone);
          }

          if (cylinder) {
            this.cylinder.setColor(cylinder);
          }

          if (torus) {
            setColor(this.torus, this.cone.material.color);
          }

          if (circleCone) {
            this.circleCone.setColor(circleCone);
          }
        }

        setAlpha({ cone, cylinder, torus, circleCone }) {
          if (cone) {
            this.cone.setAlpha(cone);
          }

          if (cylinder) {
            this.cylinder.setAlpha(cylinder);
          }

          if (torus) {
            this.torus.material.opacity = Math$1.clamp(torus, 0, 1);
          }

          if (circleCone) {
            this.circleCone.setAlpha(circleCone);
          }
        }
      }

      class Wrench extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_WRENCH) {
          super(ros, topicName, MESSAGE_TYPE_WRENCHSTAMPED, options);
          this.object = new Group$1();
          this.primitive1 = null;
          this.primitive2 = null;
          this.updateOptions({
            ...DEFAULT_OPTIONS_WRENCH,
            ...options,
          });
        }

        static getNewPrimitive(options) {
          const { type } = options;
          let newObject = null;

          switch (type) {
            case WRENCH_OBJECT_TYPES.arrow:
              newObject = new Arrow();
              break;
            case WRENCH_OBJECT_TYPES.arrowWithCircle:
              newObject = new ArrowWithCircle();
              break;
          }
          return newObject;
        }

        updatePrimitive(primitive) {
          if (primitive && primitive.type !== this.options.type) {
            this.object.remove(primitive);
            primitive = null;
          }

          if (!primitive) {
            primitive = Wrench.getNewPrimitive(this.options);
            this.object.add(primitive);
          }

          return primitive;
        }

        updateOptions(options) {
          super.updateOptions(options);
          this.options.type = WRENCH_OBJECT_TYPES.arrow;
          this.primitive1 = this.updatePrimitive(this.primitive1);
          this.options.type = WRENCH_OBJECT_TYPES.arrowWithCircle;
          this.primitive2 = this.updatePrimitive(this.primitive2);

          const forceOptions = {
            color: this.options.forceColor,
            alpha: this.options.alpha,
            headLength: this.options.headLength * this.options.forceArrowScale,
            headRadius: this.options.headRadius * this.options.arrowWidth,
            shaftLength: this.options.shaftLength * this.options.forceArrowScale,
            shaftRadius: this.options.shaftRadius * this.options.arrowWidth,
            type: WRENCH_OBJECT_TYPES.arrow,
          };

          const torqueOptions = {
            color: this.options.torqueColor,
            alpha: this.options.alpha,
            arc: this.options.arc,
            circleConeRadius: this.options.circleConeRadius * this.options.arrowWidth,
            circleConeLength: this.options.circleConeLength * this.options.arrowWidth,
            circleRadius: this.options.circleRadius * this.options.torqueArrowScale,
            headLength: this.options.headLength * this.options.torqueArrowScale,
            headRadius: this.options.headRadius * this.options.arrowWidth,
            shaftLength: this.options.shaftLength * this.options.torqueArrowScale,
            shaftRadius: this.options.shaftRadius * this.options.arrowWidth,
            tube: this.options.tube * this.options.arrowWidth,
            tubularSegments: this.options.tubularSegments,
            type: WRENCH_OBJECT_TYPES.arrowWithCircle,
          };

          setObjectDimension(this.primitive1, forceOptions);
          setObjectDimension(this.primitive2, torqueOptions);
        }

        update(message) {
          super.update(message);
          const {
            wrench: { force, torque },
          } = message;
          const translationVector = new Vector3(1, 0, 0);
          const forceVector = new Vector3(force.x, force.y, force.z).normalize();
          const torqueVector = new Vector3(torque.x, torque.y, torque.z).normalize();
          const forceQuaternion = new Quaternion().setFromUnitVectors(
            translationVector,
            forceVector,
          );
          const torqueQuaternion = new Quaternion().setFromUnitVectors(
            translationVector,
            torqueVector,
          );
          this.primitive1.setTransform({
            translation: { y: 0, x: 0, z: 0 },
            rotation: forceQuaternion,
          });
          this.primitive2.setTransform({
            translation: { y: 0, x: 0, z: 0 },
            rotation: torqueQuaternion,
          });
        }
      }

      class Polygon extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_POLYGON) {
          super(ros, topicName, MESSAGE_TYPE_POLYGONSTAMPED, {
            ...DEFAULT_OPTIONS_POLYGON,
            ...options,
          });
          this.object = new Group();
          this.line = new Line(null, true);
          this.updateOptions({
            ...DEFAULT_OPTIONS_POLYGON,
            ...options,
          });
        }

        updateOptions(options) {
          super.updateOptions(options);
          const { alpha, color } = this.options;
          this.line.setColor(new Color(color));
          this.line.setAlpha(alpha);
        }

        update(message) {
          super.update(message);
          const {
            polygon: { points },
          } = message;
          points.push(points[0]);
          this.line.updatePoints(points);
          this.object.add(this.line);
        }
      }

      class Text extends Mesh {
        constructor(text) {
          super();
          const loader = new FontLoader();

          loader.load(
            'https://raw.githubusercontent.com/mrdoob/three.js/dev/examples/fonts/helvetiker_regular.typeface.json',
            font => {
              this.geometry = new TextGeometry(text, {
                font,
                size: 0.05,
                height: 0.005,
                curveSegments: 12,
                bevelEnabled: false,
                bevelThickness: 10,
                bevelSize: 8,
                bevelSegments: 5,
              });
              this.material = new MeshBasicMaterial({ color: 0xdddddd });
            },
          );

          this.rotateX(Math.PI / 2);
          this.rotateY(Math.PI);
        }
      }

      class TfFrame extends Group {
        constructor(frameId) {
          super();
          this.add(new Axes(0.015, 0.25));
          const textObject = new Text(frameId);
          textObject
            .rotateY(Math.PI)
            .translateX(0.03)
            .translateY(0.03);
          this.add(textObject);
          this.arrow = new Arrow();
          this.arrow.setHeadDimensions({
            length: (DEFAULT_CONE_HEIGHT * 0.3) / 2,
            radius: (DEFAULT_CONE_RADIUS * 0.1) / 2,
          });
          this.arrow.setShaftDimensions({
            length: DEFAULT_CYLINDER_HEIGHT * 0.85,
            radius: (DEFAULT_CYLINDER_RADIUS * 0.05) / 6,
          });
          this.arrow.setColor({
            cone: new Color('#FF1493'),
            cylinder: new Color(DEFAULT_COLOR_ARROW),
          });
          this.add(this.arrow);
          this.name = TfFrame.getName(frameId);
        }

        static getName(frameId) {
          return `tf-${frameId}`;
        }
      }

      class Tf extends Core {
        constructor(ros, topics, options = DEFAULT_OPTIONS_TF) {
          super(ros, topics, null, {
            ...DEFAULT_OPTIONS_TF,
            ...options,
          });
          this.object = new Group$1();
          this.object.name = 'test';
        }

        update(message) {
          const { transforms } = message;
          transforms.forEach(
            ({
              header: { frame_id: parentFrameId },
              child_frame_id: childFrameId,
              transform,
            }) => {
              const [childFrame, parentFrame] = [
                this.getFrameOrCreate(childFrameId),
                this.getFrameOrCreate(parentFrameId),
              ];

              parentFrame.add(childFrame);
              childFrame.setTransform(transform);

              if (childFrame.position.length() < 0.1) {
                childFrame.arrow.visible = false;
              } else {
                childFrame.arrow.lookAt(parentFrame.getWorldPosition(new Vector3()));
                childFrame.arrow.rotateY(-Math.PI / 2);
                childFrame.arrow.visible = true;

                const arrowConeLength = childFrame.arrow.cone.scale.y;
                childFrame.arrow.setShaftDimensions({
                  length: childFrame.position.length() - arrowConeLength,
                });
              }
            },
          );

          this.object.children.forEach(child => {
            child.arrow.visible = false;
          });
        }

        getFrameOrCreate(frameId) {
          const existingFrame = this.object.getObjectByName(TfFrame.getName(frameId));
          if (existingFrame) {
            return existingFrame;
          }

          const newFrame = new TfFrame(frameId);
          this.object.add(newFrame);
          return newFrame;
        }
      }

      var commonjsGlobal = typeof globalThis !== 'undefined' ? globalThis : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

      function unwrapExports (x) {
      	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
      }

      function createCommonjsModule(fn, module) {
      	return module = { exports: {} }, fn(module, module.exports), module.exports;
      }

      var URDFLoader = createCommonjsModule(function (module, exports) {
      (function (global, factory) {
         module.exports = factory(three, STLLoader, ColladaLoader, threeMeshBvh) ;
      }(commonjsGlobal, function (THREE, STLLoader_js, ColladaLoader_js, threeMeshBvh) {
        function _typeof(obj) {
          if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
            _typeof = function (obj) {
              return typeof obj;
            };
          } else {
            _typeof = function (obj) {
              return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
            };
          }

          return _typeof(obj);
        }

        function _classCallCheck(instance, Constructor) {
          if (!(instance instanceof Constructor)) {
            throw new TypeError("Cannot call a class as a function");
          }
        }

        function _defineProperties(target, props) {
          for (var i = 0; i < props.length; i++) {
            var descriptor = props[i];
            descriptor.enumerable = descriptor.enumerable || false;
            descriptor.configurable = true;
            if ("value" in descriptor) descriptor.writable = true;
            Object.defineProperty(target, descriptor.key, descriptor);
          }
        }

        function _createClass(Constructor, protoProps, staticProps) {
          if (protoProps) _defineProperties(Constructor.prototype, protoProps);
          if (staticProps) _defineProperties(Constructor, staticProps);
          return Constructor;
        }

        function _inherits(subClass, superClass) {
          if (typeof superClass !== "function" && superClass !== null) {
            throw new TypeError("Super expression must either be null or a function");
          }

          subClass.prototype = Object.create(superClass && superClass.prototype, {
            constructor: {
              value: subClass,
              writable: true,
              configurable: true
            }
          });
          if (superClass) _setPrototypeOf(subClass, superClass);
        }

        function _getPrototypeOf(o) {
          _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
            return o.__proto__ || Object.getPrototypeOf(o);
          };
          return _getPrototypeOf(o);
        }

        function _setPrototypeOf(o, p) {
          _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
            o.__proto__ = p;
            return o;
          };

          return _setPrototypeOf(o, p);
        }

        function _assertThisInitialized(self) {
          if (self === void 0) {
            throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
          }

          return self;
        }

        function _possibleConstructorReturn(self, call) {
          if (call && (typeof call === "object" || typeof call === "function")) {
            return call;
          }

          return _assertThisInitialized(self);
        }

        function _superPropBase(object, property) {
          while (!Object.prototype.hasOwnProperty.call(object, property)) {
            object = _getPrototypeOf(object);
            if (object === null) break;
          }

          return object;
        }

        function _get(target, property, receiver) {
          if (typeof Reflect !== "undefined" && Reflect.get) {
            _get = Reflect.get;
          } else {
            _get = function _get(target, property, receiver) {
              var base = _superPropBase(target, property);

              if (!base) return;
              var desc = Object.getOwnPropertyDescriptor(base, property);

              if (desc.get) {
                return desc.get.call(receiver);
              }

              return desc.value;
            };
          }

          return _get(target, property, receiver || target);
        }

        function _slicedToArray(arr, i) {
          return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
        }

        function _toConsumableArray(arr) {
          return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
        }

        function _arrayWithoutHoles(arr) {
          if (Array.isArray(arr)) {
            for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

            return arr2;
          }
        }

        function _arrayWithHoles(arr) {
          if (Array.isArray(arr)) return arr;
        }

        function _iterableToArray(iter) {
          if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
        }

        function _iterableToArrayLimit(arr, i) {
          if (!(Symbol.iterator in Object(arr) || Object.prototype.toString.call(arr) === "[object Arguments]")) {
            return;
          }

          var _arr = [];
          var _n = true;
          var _d = false;
          var _e = undefined;

          try {
            for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
              _arr.push(_s.value);

              if (i && _arr.length === i) break;
            }
          } catch (err) {
            _d = true;
            _e = err;
          } finally {
            try {
              if (!_n && _i["return"] != null) _i["return"]();
            } finally {
              if (_d) throw _e;
            }
          }

          return _arr;
        }

        function _nonIterableSpread() {
          throw new TypeError("Invalid attempt to spread non-iterable instance");
        }

        function _nonIterableRest() {
          throw new TypeError("Invalid attempt to destructure non-iterable instance");
        }

        function URDFColliderClone() {
          var _proto$clone;

          var proto = Object.getPrototypeOf(this);

          for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
            args[_key] = arguments[_key];
          }

          var result = (_proto$clone = proto.clone).call.apply(_proto$clone, [this].concat(args));

          result.isURDFCollider = true;
          return result;
        }

        function makeURDFCollider(object) {
          object.isURDFCollider = true;
          object.clone = URDFColliderClone;
        }

        var URDFLink =
        /*#__PURE__*/
        function (_Object3D) {
          _inherits(URDFLink, _Object3D);

          function URDFLink() {
            var _getPrototypeOf2;

            var _this;

            _classCallCheck(this, URDFLink);

            for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
              args[_key2] = arguments[_key2];
            }

            _this = _possibleConstructorReturn(this, (_getPrototypeOf2 = _getPrototypeOf(URDFLink)).call.apply(_getPrototypeOf2, [this].concat(args)));
            _this.isURDFLink = true;
            _this.type = 'URDFLink';
            _this.urdfNode = null;
            return _this;
          }

          _createClass(URDFLink, [{
            key: "copy",
            value: function copy(source, recursive) {
              _get(_getPrototypeOf(URDFLink.prototype), "copy", this).call(this, source, recursive);

              this.urdfNode = source.urdfNode;
              return this;
            }
          }]);

          return URDFLink;
        }(THREE.Object3D);

        var URDFJoint =
        /*#__PURE__*/
        function (_Object3D2) {
          _inherits(URDFJoint, _Object3D2);

          _createClass(URDFJoint, [{
            key: "jointType",
            get: function get() {
              return this._jointType;
            },
            set: function set(v) {
              if (this.jointType === v) return;
              this._jointType = v;

              switch (v) {
                case 'fixed':
                case 'continuous':
                case 'revolute':
                case 'prismatic':
                  this.jointValue = 0;
                  break;

                case 'planar':
                  this.jointValue = new Array(2).fill(0);
                  break;

                case 'floating':
                  this.jointValue = new Array(6).fill(0);
                  break;
              }
            }
          }, {
            key: "angle",
            get: function get() {
              return this.jointValue;
            }
          }]);

          function URDFJoint() {
            var _getPrototypeOf3;

            var _this2;

            _classCallCheck(this, URDFJoint);

            for (var _len3 = arguments.length, args = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
              args[_key3] = arguments[_key3];
            }

            _this2 = _possibleConstructorReturn(this, (_getPrototypeOf3 = _getPrototypeOf(URDFJoint)).call.apply(_getPrototypeOf3, [this].concat(args)));
            _this2.isURDFJoint = true;
            _this2.type = 'URDFJoint';
            _this2.urdfNode = null;
            _this2.jointValue = null;
            _this2.jointType = 'fixed';
            _this2.axis = null;
            _this2.limit = {
              lower: 0,
              upper: 0
            };
            _this2.ignoreLimits = false;
            _this2.origPosition = null;
            _this2.origQuaternion = null;
            return _this2;
          }
          /* Overrides */


          _createClass(URDFJoint, [{
            key: "copy",
            value: function copy(source, recursive) {
              _get(_getPrototypeOf(URDFJoint.prototype), "copy", this).call(this, source, recursive);

              this.urdfNode = source.urdfNode;
              this.jointType = source.jointType;
              this.axis = source.axis ? source.axis.clone() : null;
              this.limit.lower = source.limit.lower;
              this.limit.upper = source.limit.upper;
              this.ignoreLimits = false;
              this.jointValue = Array.isArray(source.jointValue) ? _toConsumableArray(source.jointValue) : source.jointValue;
              this.origPosition = source.origPosition ? source.origPosition.clone() : null;
              this.origQuaternion = source.origQuaternion ? source.origQuaternion.clone() : null;
              return this;
            }
            /* Public Functions */

          }, {
            key: "setAngle",
            value: function setAngle() {
              return this.setOffset.apply(this, arguments);
            }
          }, {
            key: "setOffset",
            value: function setOffset() {
              for (var _len4 = arguments.length, values = new Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
                values[_key4] = arguments[_key4];
              }

              values = values.map(function (v) {
                return parseFloat(v);
              });

              if (!this.origPosition || !this.origQuaternion) {
                this.origPosition = this.position.clone();
                this.origQuaternion = this.quaternion.clone();
              }

              switch (this.jointType) {
                case 'fixed':
                  {
                    break;
                  }

                case 'continuous':
                case 'revolute':
                  {
                    var angle = values[0];
                    if (angle == null) break;
                    if (angle === this.jointValue) break;

                    if (!this.ignoreLimits && this.jointType === 'revolute') {
                      angle = Math.min(this.limit.upper, angle);
                      angle = Math.max(this.limit.lower, angle);
                    } // FromAxisAngle seems to rotate the opposite of the
                    // expected angle for URDF, so negate it here


                    var delta = new THREE.Quaternion().setFromAxisAngle(this.axis, angle);
                    this.quaternion.multiplyQuaternions(this.origQuaternion, delta);
                    this.jointValue = angle;
                    this.matrixWorldNeedsUpdate = true;
                    break;
                  }

                case 'prismatic':
                  {
                    var _angle = values[0];
                    if (_angle == null) break;
                    if (_angle === this.jointValue) break;

                    if (!this.ignoreLimits) {
                      _angle = Math.min(this.limit.upper, _angle);
                      _angle = Math.max(this.limit.lower, _angle);
                    }

                    this.position.copy(this.origPosition);
                    this.position.addScaledVector(this.axis, _angle);
                    this.jointValue = _angle;
                    this.worldMatrixNeedsUpdate = true;
                    break;
                  }

                case 'floating':
                case 'planar':
                  // TODO: Support these joint types
                  console.warn("'".concat(this.jointType, "' joint not yet supported"));
              }

              return this.jointValue;
            }
          }]);

          return URDFJoint;
        }(THREE.Object3D);

        var URDFRobot =
        /*#__PURE__*/
        function (_URDFLink) {
          _inherits(URDFRobot, _URDFLink);

          function URDFRobot() {
            var _getPrototypeOf4;

            var _this3;

            _classCallCheck(this, URDFRobot);

            for (var _len5 = arguments.length, args = new Array(_len5), _key5 = 0; _key5 < _len5; _key5++) {
              args[_key5] = arguments[_key5];
            }

            _this3 = _possibleConstructorReturn(this, (_getPrototypeOf4 = _getPrototypeOf(URDFRobot)).call.apply(_getPrototypeOf4, [this].concat(args)));
            _this3.isURDFRobot = true;
            _this3.urdfNode = null;
            _this3.urdfRobotNode = null;
            _this3.robotName = null;
            _this3.links = null;
            _this3.joints = null;
            return _this3;
          }

          _createClass(URDFRobot, [{
            key: "copy",
            value: function copy(source, recursive) {
              var _this4 = this;

              _get(_getPrototypeOf(URDFRobot.prototype), "copy", this).call(this, source, recursive);

              this.urdfRobotNode = source.urdfRobotNode;
              this.robotName = source.robotName;
              this.links = {};
              this.joints = {};
              this.traverse(function (c) {
                if (c.isURDFJoint && c.name in source.joints) {
                  _this4.joints[c.name] = c;
                }

                if (c.isURDFLink && c.name in source.links) {
                  _this4.links[c.name] = c;
                }
              });
              return this;
            }
          }, {
            key: "setAngle",
            value: function setAngle(jointName) {
              var joint = this.joints[jointName];

              if (joint) {
                for (var _len6 = arguments.length, angle = new Array(_len6 > 1 ? _len6 - 1 : 0), _key6 = 1; _key6 < _len6; _key6++) {
                  angle[_key6 - 1] = arguments[_key6];
                }

                return joint.setAngle.apply(joint, angle);
              }

              return null;
            }
          }, {
            key: "setAngles",
            value: function setAngles(angles) {
              // TODO: How to handle other, multi-dimensional joint types?
              for (var name in angles) {
                this.setAngle(name, angles[name]);
              }
            }
          }]);

          return URDFRobot;
        }(URDFLink);

        /*
        Reference coordinate frames for THREE.js and ROS.
        Both coordinate systems are right handed so the URDF is instantiated without
        frame transforms. The resulting model can be rotated to rectify the proper up,
        right, and forward directions

        THREE.js
           Y
           |
           |
           .-----X
         ／
        Z

        ROS URDf
               Z
               |   X
               | ／
         Y-----.

        */

        var tempQuaternion = new THREE.Quaternion();
        var tempEuler = new THREE.Euler(); // take a vector "x y z" and process it into
        // an array [x, y, z]

        function processTuple(val) {
          if (!val) return [0, 0, 0];
          return val.trim().split(/\s+/g).map(function (num) {
            return parseFloat(num);
          });
        } // applies a rotation a threejs object in URDF order


        function applyRotation(obj, rpy) {
          var additive = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
          // if additive is true the rotation is applied in
          // addition to the existing rotation
          if (!additive) obj.rotation.set(0, 0, 0);
          tempEuler.set(rpy[0], rpy[1], rpy[2], 'ZYX');
          tempQuaternion.setFromEuler(tempEuler);
          tempQuaternion.multiply(obj.quaternion);
          obj.quaternion.copy(tempQuaternion);
        }
        /* URDFLoader Class */
        // Loads and reads a URDF file into a THREEjs Object3D format


        var URDFLoader =
        /*#__PURE__*/
        function () {
          function URDFLoader(manager) {
            _classCallCheck(this, URDFLoader);

            this.manager = manager || THREE.DefaultLoadingManager;
          }
          /* Public API */
          // urdf:    The path to the URDF within the package OR absolute
          // onComplete:      Callback that is passed the model once loaded


          _createClass(URDFLoader, [{
            key: "load",
            value: function load(urdf, onComplete, onProgress, onError, options) {
              var _this = this;

              // Check if a full URI is specified before
              // prepending the package info
              var manager = this.manager;
              var workingPath = THREE.LoaderUtils.extractUrlBase(urdf);
              var urdfPath = this.manager.resolveURL(urdf);
              var errors = {};

              var managerOnErrorDefault = function managerOnErrorDefault() {};

              var managerOnProgressDefault = function managerOnProgressDefault() {};

              var managerOnLoadDefault = function managerOnLoadDefault() {};

              var model;

              if (manager.onError) {
                managerOnErrorDefault = manager.onProgress.bind(manager);
              }

              if (manager.onProgress) {
                managerOnProgressDefault = manager.onProgress.bind(manager);
              }

              if (manager.onLoad) {
                managerOnLoadDefault = manager.onLoad.bind(manager);
              }

              manager.onError = function (url) {
                errors[url] = 'Error in loading resource';

                if (onError) {
                  onError(url);
                }

                managerOnErrorDefault(url);
              };

              manager.onProgress = function (url, itemsLoaded, itemsTotal) {
                if (onProgress) {
                  onProgress(url, itemsLoaded, itemsTotal);
                }

                managerOnProgressDefault(url, itemsLoaded, itemsTotal);
              };

              manager.onLoad = function () {
                if (onComplete) {
                  var partialErrors = Object.keys(errors).length === 0 ? undefined : errors;
                  onComplete(model, partialErrors);
                }

                managerOnLoadDefault();
              };

              options = Object.assign({
                workingPath: workingPath
              }, options);
              manager.itemStart(urdfPath);
              fetch(urdfPath, options.fetchOptions).then(function (res) {
                return res.text();
              }).then(function (data) {
                model = _this.parse(data, options);
                manager.itemEnd(urdfPath);
              })["catch"](function (e) {
                console.error('URDFLoader: Error parsing file.', e);
                manager.itemError(urdfPath);
                manager.itemEnd(urdfPath);
              });
            }
          }, {
            key: "parse",
            value: function parse(content) {
              var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
              var packages = options.packages || '';
              var loadMeshCb = options.loadMeshCb || this.defaultMeshLoader.bind(this);
              var workingPath = options.workingPath || '';
              var parseVisual = 'parseVisual' in options ? options.parseVisual : true;
              var parseCollision = options.parseCollision || false;
              var manager = this.manager;
              var linkMap = {};
              var jointMap = {};
              var materialMap = {}; // Resolves the path of mesh files

              function resolvePath(path) {
                if (!/^package:\/\//.test(path)) {
                  return workingPath ? workingPath + path : path;
                } // Remove "package://" keyword and split meshPath at the first slash


                var _path$replace$split = path.replace(/^package:\/\//, '').split(/\/(.+)/),
                    _path$replace$split2 = _slicedToArray(_path$replace$split, 2),
                    targetPkg = _path$replace$split2[0],
                    relPath = _path$replace$split2[1];

                if (typeof packages === 'string') {
                  // "pkg" is one single package
                  if (packages.endsWith(targetPkg)) {
                    // "pkg" is the target package
                    return packages + '/' + relPath;
                  } else {
                    // Assume "pkg" is the target package's parent directory
                    return packages + '/' + targetPkg + '/' + relPath;
                  }
                } else if (_typeof(packages) === 'object') {
                  // "pkg" is a map of packages
                  if (targetPkg in packages) {
                    return packages[targetPkg] + '/' + relPath;
                  } else {
                    console.error("URDFLoader : ".concat(targetPkg, " not found in provided package list."));
                    return null;
                  }
                }
              } // Process the URDF text format


              function processUrdf(data) {
                var parser = new DOMParser();
                var urdf = parser.parseFromString(data, 'text/xml');

                var children = _toConsumableArray(urdf.children);

                var robotNode = children.filter(function (c) {
                  return c.nodeName === 'robot';
                }).pop();
                return processRobot(robotNode);
              } // Process the <robot> node


              function processRobot(robot) {
                var robotNodes = _toConsumableArray(robot.children);

                var links = robotNodes.filter(function (c) {
                  return c.nodeName.toLowerCase() === 'link';
                });
                var joints = robotNodes.filter(function (c) {
                  return c.nodeName.toLowerCase() === 'joint';
                });
                var materials = robotNodes.filter(function (c) {
                  return c.nodeName.toLowerCase() === 'material';
                });
                var obj = new URDFRobot();
                obj.robotName = robot.getAttribute('name');
                obj.urdfRobotNode = robot; // Create the <material> map

                materials.forEach(function (m) {
                  var name = m.getAttribute('name');
                  materialMap[name] = processMaterial(m);
                }); // Create the <link> map

                links.forEach(function (l) {
                  var name = l.getAttribute('name');
                  var isRoot = robot.querySelector("child[link=\"".concat(name, "\"]")) === null;
                  linkMap[name] = processLink(l, isRoot ? obj : null);
                }); // Create the <joint> map

                joints.forEach(function (j) {
                  var name = j.getAttribute('name');
                  jointMap[name] = processJoint(j);
                });
                obj.joints = jointMap;
                obj.links = linkMap;
                return obj;
              } // Process joint nodes and parent them


              function processJoint(joint) {
                var children = _toConsumableArray(joint.children);

                var jointType = joint.getAttribute('type');
                var obj = new URDFJoint();
                obj.urdfNode = joint;
                obj.name = joint.getAttribute('name');
                obj.jointType = jointType;
                var parent = null;
                var child = null;
                var xyz = [0, 0, 0];
                var rpy = [0, 0, 0]; // Extract the attributes

                children.forEach(function (n) {
                  var type = n.nodeName.toLowerCase();

                  if (type === 'origin') {
                    xyz = processTuple(n.getAttribute('xyz'));
                    rpy = processTuple(n.getAttribute('rpy'));
                  } else if (type === 'child') {
                    child = linkMap[n.getAttribute('link')];
                  } else if (type === 'parent') {
                    parent = linkMap[n.getAttribute('link')];
                  } else if (type === 'limit') {
                    obj.limit.lower = parseFloat(n.getAttribute('lower') || obj.limit.lower);
                    obj.limit.upper = parseFloat(n.getAttribute('upper') || obj.limit.upper);
                  }
                }); // Join the links

                parent.add(obj);
                obj.add(child);
                applyRotation(obj, rpy);
                obj.position.set(xyz[0], xyz[1], xyz[2]); // Set up the rotate function

                var axisNode = children.filter(function (n) {
                  return n.nodeName.toLowerCase() === 'axis';
                })[0];

                if (axisNode) {
                  var axisXYZ = axisNode.getAttribute('xyz').split(/\s+/g).map(function (num) {
                    return parseFloat(num);
                  });
                  obj.axis = new THREE.Vector3(axisXYZ[0], axisXYZ[1], axisXYZ[2]);
                  obj.axis.normalize();
                }

                return obj;
              } // Process the <link> nodes


              function processLink(link) {
                var target = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

                if (target === null) {
                  target = new URDFLink();
                }

                var children = _toConsumableArray(link.children);

                target.name = link.getAttribute('name');
                target.urdfNode = link;

                if (parseVisual) {
                  var visualNodes = children.filter(function (n) {
                    return n.nodeName.toLowerCase() === 'visual';
                  });
                  visualNodes.forEach(function (vn) {
                    return processLinkElement(vn, target, materialMap);
                  });
                }

                if (parseCollision) {
                  var collisionNodes = children.filter(function (n) {
                    return n.nodeName.toLowerCase() === 'collision';
                  });
                  collisionNodes.forEach(function (vn) {
                    return processLinkElement(vn, target);
                  });
                }

                return target;
              }

              function processMaterial(node) {
                var matNodes = _toConsumableArray(node.children);

                var material = new THREE.MeshPhongMaterial();
                material.name = node.getAttribute('name') || '';
                matNodes.forEach(function (n) {
                  var type = n.nodeName.toLowerCase();

                  if (type === 'color') {
                    var rgba = n.getAttribute('rgba').split(/\s/g).map(function (v) {
                      return parseFloat(v);
                    });
                    material.color.setRGB(rgba[0], rgba[1], rgba[2]);
                    material.opacity = rgba[3];
                    material.transparent = rgba[3] < 1;
                  } else if (type === 'texture') {
                    var loader = new THREE.TextureLoader(manager);
                    var filename = n.getAttribute('filename');
                    var filePath = resolvePath(filename);
                    material.map = loader.load(filePath);
                  }
                });
                return material;
              } // Process the visual and collision nodes into meshes


              function processLinkElement(vn, linkObj) {
                var materialMap = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
                var isCollisionNode = vn.nodeName.toLowerCase() === 'collision';
                var xyz = [0, 0, 0];
                var rpy = [0, 0, 0];
                var scale = [1, 1, 1];

                var children = _toConsumableArray(vn.children);

                var material = null;
                var primitiveModel = null; // get the material first

                var materialNode = children.filter(function (n) {
                  return n.nodeName.toLowerCase() === 'material';
                })[0];

                if (materialNode) {
                  var name = materialNode.getAttribute('name');

                  if (name && name in materialMap) {
                    material = materialMap[name];
                  } else {
                    material = processMaterial(materialNode);
                  }
                } else {
                  material = new THREE.MeshPhongMaterial();
                }

                children.forEach(function (n) {
                  var type = n.nodeName.toLowerCase();

                  if (type === 'geometry') {
                    var geoType = n.children[0].nodeName.toLowerCase();

                    if (geoType === 'mesh') {
                      var filename = n.children[0].getAttribute('filename');
                      var filePath = resolvePath(filename); // file path is null if a package directory is not provided.

                      if (filePath !== null) {
                        var scaleAttr = n.children[0].getAttribute('scale');
                        if (scaleAttr) scale = processTuple(scaleAttr);
                        loadMeshCb(filePath, manager, function (obj, err) {
                          if (err) {
                            console.error('URDFLoader: Error loading mesh.', err);
                          } else if (obj) {
                            if (obj instanceof THREE.Mesh) {
                              obj.material = material;
                              obj.raycast = threeMeshBvh.acceleratedRaycast;
                              obj.geometry.boundsTree = new threeMeshBvh.MeshBVH(obj.geometry);
                            }

                            linkObj.add(obj);
                            obj.position.set(xyz[0], xyz[1], xyz[2]);
                            obj.rotation.set(0, 0, 0); // multiply the existing scale by the scale components because
                            // the loaded model could have important scale values already applied
                            // to the root. Collada files, for example, can load in with a scale
                            // to convert the model units to meters.

                            obj.scale.x *= scale[0];
                            obj.scale.y *= scale[1];
                            obj.scale.z *= scale[2];
                            applyRotation(obj, rpy);

                            if (isCollisionNode) {
                              makeURDFCollider(obj);
                            }
                          }
                        });
                      }
                    } else if (geoType === 'box') {
                      primitiveModel = new THREE.Mesh();
                      primitiveModel.geometry = new THREE.BoxBufferGeometry(1, 1, 1);
                      primitiveModel.material = material;
                      primitiveModel.raycast = threeMeshBvh.acceleratedRaycast;
                      primitiveModel.geometry.boundsTree = new threeMeshBvh.MeshBVH(primitiveModel.geometry);
                      var size = processTuple(n.children[0].getAttribute('size'));
                      linkObj.add(primitiveModel);
                      primitiveModel.scale.set(size[0], size[1], size[2]);

                      if (isCollisionNode) {
                        makeURDFCollider(primitiveModel);
                      }
                    } else if (geoType === 'sphere') {
                      primitiveModel = new THREE.Mesh();
                      primitiveModel.geometry = new THREE.SphereBufferGeometry(1, 30, 30);
                      primitiveModel.material = material;
                      primitiveModel.raycast = threeMeshBvh.acceleratedRaycast;
                      primitiveModel.geometry.boundsTree = new threeMeshBvh.MeshBVH(primitiveModel.geometry);
                      var radius = parseFloat(n.children[0].getAttribute('radius')) || 0;
                      primitiveModel.scale.set(radius, radius, radius);
                      linkObj.add(primitiveModel);

                      if (isCollisionNode) {
                        makeURDFCollider(primitiveModel);
                      }
                    } else if (geoType === 'cylinder') {
                      primitiveModel = new THREE.Mesh();
                      primitiveModel.geometry = new THREE.CylinderBufferGeometry(1, 1, 1, 30);
                      primitiveModel.material = material;
                      primitiveModel.raycast = threeMeshBvh.acceleratedRaycast;
                      primitiveModel.geometry.boundsTree = new threeMeshBvh.MeshBVH(primitiveModel.geometry);

                      var _radius = parseFloat(n.children[0].getAttribute('radius')) || 0;

                      var length = parseFloat(n.children[0].getAttribute('length')) || 0;
                      primitiveModel.scale.set(_radius, length, _radius);
                      primitiveModel.rotation.set(Math.PI / 2, 0, 0);
                      linkObj.add(primitiveModel);

                      if (isCollisionNode) {
                        makeURDFCollider(primitiveModel);
                      }
                    }
                  } else if (type === 'origin') {
                    xyz = processTuple(n.getAttribute('xyz'));
                    rpy = processTuple(n.getAttribute('rpy'));
                  }
                }); // apply the position and rotation to the primitive geometry after
                // the fact because it's guaranteed to have been scraped from the child
                // nodes by this point

                if (primitiveModel) {
                  applyRotation(primitiveModel, rpy, true);
                  primitiveModel.position.set(xyz[0], xyz[1], xyz[2]);
                }
              }

              return processUrdf(content);
            } // Default mesh loading function

          }, {
            key: "defaultMeshLoader",
            value: function defaultMeshLoader(path, manager, done) {
              if (/\.stl$/i.test(path)) {
                var loader = new STLLoader_js.STLLoader(manager);
                loader.load(path, function (geom) {
                  var mesh = new THREE.Mesh(geom, new THREE.MeshPhongMaterial());
                  done(mesh);
                });
              } else if (/\.dae$/i.test(path)) {
                var _loader = new ColladaLoader_js.ColladaLoader(manager);

                _loader.load(path, function (dae) {
                  return done(dae.scene);
                });
              } else {
                console.warn("URDFLoader: Could not load model at ".concat(path, ".\nNo loader available"));
              }
            }
          }]);

          return URDFLoader;
        }();

        return URDFLoader;

      }));

      });

      var URDFLoader$1 = unwrapExports(URDFLoader);

      const excludedObjects = [
        'PerspectiveCamera',
        'OrthographicCamera',
        'AmbientLight',
        'DirectionalLight',
        'HemisphereLight',
        'Light',
        'RectAreaLight',
        'SpotLight',
        'PointLight',
      ];

      const removeExcludedObjects = mesh => {
        const objectArray = [mesh];
        while (Object.keys(objectArray).length > 0) {
          const currentItem = objectArray.shift();
          currentItem.children.forEach(child => {
            if (!child) {
              return;
            }
            if (excludedObjects.indexOf(child.type) > -1) {
              const { parent } = child;
              parent.children = parent.children.filter(c => c !== child);
            } else {
              objectArray.push(child);
            }
          });
        }
      };

      class RobotModel extends URDFLoader$1 {
        constructor(ros, paramName, options = DEFAULT_OPTIONS_ROBOTMODEL) {
          super(DefaultLoadingManager);
          const { packages } = options;
          this.param = new ROSLIB.Param({
            ros,
            name: paramName,
          });
          this.object = new Group();
          this.packages = packages || {};
          this.updateOptions({
            ...DEFAULT_OPTIONS_ROBOTMODEL,
            ...options,
          });

          this.defaultLoadMeshCallback = this.defaultLoadMeshCallback.bind(this);
        }

        static onComplete(object) {
          removeExcludedObjects(object);
        }

        updateOptions() {}

        load(onComplete = RobotModel.onComplete, options = {}) {
          this.param.get(robotString => {
            this.loadRobot(robotString, onComplete, options);
          });
        }

        loadRobot(robotString, onComplete = RobotModel.onComplete, options = {}) {
          const robotModel = super.parse(robotString, {
            packages: this.packages,
            loadMeshCb: options.loadMeshCb || this.defaultLoadMeshCallback,
            fetchOptions: { mode: 'cors', credentials: 'same-origin' },
            ...options,
          });
          this.object.add(robotModel);
          this.object.name = robotModel.robotName;

          onComplete(this.object);
        }

        defaultLoadMeshCallback(path, ext, done) {
          super.defaultMeshLoader(path, ext, mesh => {
            removeExcludedObjects(mesh);
            done(mesh);
          });
        }

        getPackages(onComplete) {
          this.param.get(robotString => {
            const parser = new DOMParser();
            const urdf = parser.parseFromString(robotString, 'text/xml');
            const packages = [...urdf.querySelectorAll('mesh')].map(mesh => {
              const [targetPkg] = mesh
                .getAttribute('filename')
                .replace(/^package:\/\//, '')
                .split(/\/(.+)/);
              return targetPkg;
            });
            onComplete([...new Set(packages)]);
          });
        }

        destroy() {
          if (this.object.parent) {
            this.object.parent.remove(this.object);
          }
        }

        hide() {
          this.object.visible = false;
        }

        show() {
          this.object.visible = true;
        }
      }

      class Sphere extends Mesh {
        constructor() {
          super();
          // Radius handled through scale
          this.geometry = new SphereGeometry(
            1,
            DEFAULT_RADIAL_SEGMENTS,
            DEFAULT_RADIAL_SEGMENTS,
          );
          this.material = new MeshStandardMaterial();
          this.material.transparent = true;
        }
      }

      class Point extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_POINT) {
          super(ros, topicName, MESSAGE_TYPE_POINTSTAMPED, options);
          this.object = new Group();
          this.sphere = new Sphere();
          this.object.add(this.sphere);
          this.updateOptions({
            ...DEFAULT_OPTIONS_POINT,
            ...options,
          });
        }

        updateOptions(options) {
          super.updateOptions(options);
          const { alpha, color, radius } = this.options;
          this.sphere.setColor(color);
          this.sphere.setAlpha(alpha);
          this.sphere.setScale({ x: radius, y: radius, z: radius });
        }

        update(message) {
          super.update(message);
          const {
            point: { x, y, z },
          } = message;
          this.object.position.set(x, y, z);
        }
      }

      const getAccessorForDataType = (dataView, dataType) => {
        switch (dataType) {
          case POINT_FIELD_DATATYPES.INT8: {
            return dataView.getInt8;
          }
          case POINT_FIELD_DATATYPES.UINT8: {
            return dataView.getUint8;
          }
          case POINT_FIELD_DATATYPES.INT16: {
            return dataView.getInt16;
          }
          case POINT_FIELD_DATATYPES.UINT16: {
            return dataView.getUint16;
          }
          case POINT_FIELD_DATATYPES.INT32: {
            return dataView.getInt32;
          }
          case POINT_FIELD_DATATYPES.UINT32: {
            return dataView.getUint32;
          }
          case POINT_FIELD_DATATYPES.FLOAT64: {
            return dataView.getFloat64;
          }
          case POINT_FIELD_DATATYPES.FLOAT32:
          default: {
            return dataView.getFloat32;
          }
        }
      };

      const updateGeometryAttribute = (geometry, attribute, data, l) => {
        geometry.attributes[attribute].updateRange = {
          offset: 0,
          count: l * 3,
        };
        geometry.attributes[attribute].count = l;
        geometry.attributes[attribute].array = data;
        geometry.attributes[attribute].needsUpdate = true;
      };

      class PCLDecoder {
        static decode(message, colorChannel, useRainbow) {
          const n = message.height * message.width;
          const positions = new Float32Array(3 * n);
          const colors = new Float32Array(3 * n);
          const normals = new Float32Array(3 * n);

          const uint8Buffer = Uint8Array.from(message.data).buffer;
          const dataView = new DataView(uint8Buffer);
          const offsets = {};
          const accessor = {};

          const rgbCache = new Color();

          message.fields.forEach(f => {
            offsets[f.name] = f.offset;
            accessor[f.name] = getAccessorForDataType(dataView, message.datatype);
          });

          for (let i = 0; i < n; i++) {
            const stride = i * message.point_step;
            if (
              offsets.x !== undefined &&
              offsets.y !== undefined &&
              offsets.z !== undefined
            ) {
              positions[3 * i] = accessor.x.call(
                dataView,
                stride + offsets.x,
                !message.is_bigendian,
              );
              positions[3 * i + 1] = accessor.y.call(
                dataView,
                stride + offsets.y,
                !message.is_bigendian,
              );
              positions[3 * i + 2] = accessor.z.call(
                dataView,
                stride + offsets.z,
                !message.is_bigendian,
              );
            }

            if (
              offsets.rgb !== undefined &&
              colorChannel === POINTCLOUD_COLOR_CHANNELS.RGB
            ) {
              colors[3 * i] = dataView.getUint8(stride + offsets.rgb + 2) / 255.0;
              colors[3 * i + 1] = dataView.getUint8(stride + offsets.rgb + 1) / 255.0;
              colors[3 * i + 2] = dataView.getUint8(stride + offsets.rgb) / 255.0;
            }

            if (
              offsets.intensity !== undefined &&
              colorChannel === POINTCLOUD_COLOR_CHANNELS.INTENSITY
            ) {
              const intensity = accessor.intensity.call(
                dataView,
                stride + offsets.intensity,
                !message.is_bigendian,
              );

              const normalizedIntensity = Math.min(intensity, 360) / 360;

              if (useRainbow) {
                const rgb = rgbCache
                  .setHSL(normalizedIntensity, 1.0, 0.5)
                  .toArray()
                  .map(x => x * 255);
                colors[3 * i] = rgb[0] / 255;
                colors[3 * i + 1] = rgb[1] / 255;
                colors[3 * i + 2] = rgb[2] / 255;
              } else {
                colors[3 * i] = normalizedIntensity / 360;
                colors[3 * i + 1] = normalizedIntensity / 360;
                colors[3 * i + 2] = normalizedIntensity / 360;
              }
            }

            if (
              offsets.normal_x !== undefined &&
              offsets.normal_y !== undefined &&
              offsets.normal_z !== undefined
            ) {
              normals[3 * i] = accessor.normal_x.call(
                dataView,
                stride + offsets.normal_x,
                !message.is_bigendian,
              );
              normals[3 * i + 1] = accessor.normal_y.call(
                dataView,
                stride + offsets.normal_y,
                !message.is_bigendian,
              );
              normals[3 * i + 2] = accessor.normal_z.call(
                dataView,
                stride + offsets.normal_z,
                !message.is_bigendian,
              );
            }
          }
          return { positions, colors, normals };
        }

        static attachDecoder(module) {
          const { get_memory: getMemory, PCLDecoder: PCLDecoderWasm } = module;
          const pclDecoderModule = new PCLDecoderWasm();
          const memory = getMemory();

          PCLDecoder.decode = (message, colorChannel, useRainbow) => {
            const n = message.height * message.width;
            const offsets = {};
            const normals = new Float32Array(3 * n);

            message.fields.forEach(f => {
              offsets[f.name] = f.offset;
            });

            const memoryTypedArray = new Uint8Array(memory.buffer);

            const copyMemPtr = pclDecoderModule.get_copy_memory_ptr();

            memoryTypedArray.set(message.data, copyMemPtr);
            pclDecoderModule.compute(
              n,
              message.point_step,
              offsets.x,
              offsets.y,
              offsets.z,
              offsets.rgb || 0,
              offsets.intensity || 0,
              colorChannel === POINTCLOUD_COLOR_CHANNELS.INTENSITY,
              useRainbow,
            );
            const positionMemPointer = pclDecoderModule.get_position_memory_ptr();
            const colorMemPointer = pclDecoderModule.get_color_memory_ptr();
            const positions = new Float32Array(
              memory.buffer,
              positionMemPointer,
              3 * n,
            );
            const invalidChannel =
              (colorChannel === POINTCLOUD_COLOR_CHANNELS.INTENSITY &&
                !offsets.intensity) ||
              (colorChannel === POINTCLOUD_COLOR_CHANNELS.RGB && !offsets.rgb);
            const colors = invalidChannel
              ? new Float32Array(3 * n)
              : new Float32Array(memory.buffer, colorMemPointer, 3 * n);
            return { positions, colors, normals };
          };
        }
      }

      const editPointCloudPoints = function(message, options) {
        if (!message) {
          return {
            positions: new Float32Array(0),
            colors: new Float32Array(0),
            normals: new Float32Array(0),
          };
        }

        const { colorChannel, useRainbow } = options;

        const { colors, normals, positions } = PCLDecoder.decode(
          message,
          colorChannel,
          useRainbow,
        );

        return {
          positions,
          colors,
          normals,
        };
      };

      class PointCloud extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_POINTCLOUD) {
          super(ros, topicName, MESSAGE_TYPE_POINTCLOUD2, {
            ...DEFAULT_OPTIONS_POINTCLOUD,
            ...options,
          });
          const cloudMaterial = new PointsMaterial({
            size: this.options.size,
            vertexColors: VertexColors$1,
          });
          const geometry = new BufferGeometry();
          geometry.addAttribute(
            'position',
            new BufferAttribute(
              new Float32Array(MAX_POINTCLOUD_POINTS * 3),
              3,
            ).setDynamic(true),
          );
          geometry.addAttribute(
            'color',
            new BufferAttribute(
              new Float32Array(MAX_POINTCLOUD_POINTS * 3),
              3,
            ).setDynamic(true),
          );
          geometry.addAttribute(
            'normal',
            new BufferAttribute(
              new Float32Array(MAX_POINTCLOUD_POINTS * 3),
              3,
            ).setDynamic(true),
          );
          geometry.setDrawRange(0, 0);
          geometry.computeBoundingSphere();
          this.object = new Points$2(geometry, cloudMaterial);
          this.object.frustumCulled = false;
          this.updateOptions({
            ...DEFAULT_OPTIONS_POINTCLOUD,
            ...options,
          });

          this.importWasm();
        }

        importWasm() {
          if (typeof WebAssembly !== 'undefined') {
            module.import('pcl-decoder').then(module => {
              PCLDecoder.attachDecoder(module);
            });
          }
        }

        updatePointCloudGeometry(positions, colors, normals) {
          const { geometry, material } = this.object;
          if (
            material.size !== this.options.size &&
            !Number.isNaN(this.options.size)
          ) {
            material.size = this.options.size;
            material.needsUpdate = true;
          }
          const l = Math.min(MAX_POINTCLOUD_POINTS, Math.floor(positions.length / 3));
          geometry.setDrawRange(0, l);
          updateGeometryAttribute(geometry, 'position', positions, l);
          updateGeometryAttribute(geometry, 'color', colors, l);
          updateGeometryAttribute(geometry, 'normal', normals, l);
        }

        update(message) {
          super.update(message);
          const { colors, normals, positions } = editPointCloudPoints(
            message,
            this.options,
          );
          this.updatePointCloudGeometry(positions, colors, normals);
        }
      }

      class Cube extends Mesh {
        constructor() {
          super();
          this.geometry = new BoxGeometry();
          this.material = new MeshStandardMaterial();
        }

        setScale({ x }) {
          super.setScale({ x, y: x, z: x });
        }
      }

      class LineSegments extends LineSegments$1 {
        constructor(color, linewidth = 5) {
          super();
          this.geometry = new Geometry();
          this.material = new LineBasicMaterial({ linewidth });
          this.material.vertexColors = VertexColors;
        }

        setColor(colors) {
          setColor(this, colors);
        }

        updatePoints(points, colors) {
          this.geometry.vertices = points.map(({ x, y, z }) => new Vector3(x, y, z));
          this.geometry.verticesNeedUpdate = true;

          if (colors.length > 0) {
            this.geometry.colors = colors.map(({ r, g, b }) => new Color(r, g, b));
            this.geometry.colorsNeedUpdate = true;
          }
        }

        setTransform(transform) {
          setTransform(this, transform);
        }
      }

      class ObjectCacher {
        constructor(objectPool, Primitive) {
          this.objectPool = objectPool;
          this.Primitive = Primitive;
        }

        setObjectDimension(object, { x, y, z }, color, scale) {
          object.setColor(color);
          object.setScale(scale);
          object.position.set(x, y, z);
        }

        reusePool(points, colors, options) {
          const { scale } = options;
          const currentCount = points.length;

          for (let i = 0; i < currentCount; i++) {
            const currentChild = this.objectPool.children[i];
            this.setObjectDimension(currentChild, points[i], colors[i], scale);
          }

          for (let i = currentCount; i < this.objectPool.children.length; i++) {
            this.objectPool.children[i].visible = false;
          }
        }

        increasePool(points, colors, options) {
          const currentCount = this.objectPool.children.length;
          const { scale } = options;

          for (let i = 0; i < currentCount; i++) {
            const currentChild = this.objectPool.children[i];
            this.setObjectDimension(currentChild, points[i], colors[i], scale);
          }

          for (let i = currentCount; i < points.length; i++) {
            const sphere = new this.Primitive();
            const { x, y, z } = points[i];

            sphere.setColor(colors[i]);
            sphere.setScale(scale);
            sphere.position.set(x, y, z);
            this.objectPool.add(sphere);
          }
        }
      }

      class SphereList extends Mesh {
        constructor() {
          super();
          this.geometry = new Geometry();
          this.material = new MeshBasicMaterial();
          this.objectCacher = new ObjectCacher(this, Sphere);
        }

        updatePoints(points, colors, options = {}) {
          options.subtype = MARKER_OBJECT_TYPES.SPHERE;

          if (points.length < this.children.length) {
            this.objectCacher.reusePool(points, colors, options);
          } else {
            this.objectCacher.increasePool(points, colors, options);
          }
        }
      }

      class Points extends Points$2 {
        constructor() {
          super();
          this.geometry = new Geometry();
          this.material = new PointsMaterial({
            vertexColors: VertexColors$1,
          });
        }

        setTransform(transform) {
          setTransform(this, transform);
        }

        updatePoints(points, colors, options = {}) {
          const {
            scale: { x },
          } = options;

          this.material.size = x;
          this.geometry.vertices = points.map(
            vertex => new Vector3(vertex.x, vertex.y, vertex.z),
          );
          this.geometry.verticesNeedUpdate = true;

          if (colors.length > 0) {
            this.geometry.colors = colors.map(
              color => new Color(color.r, color.g, color.b),
            );
            this.geometry.colorsNeedUpdate = true;
          }
        }

        setScale({ x: size }) {
          this.material.size = size;
        }
      }

      class TriangleList extends Mesh {
        constructor() {
          super();
          this.geometry = new Geometry();

          this.material = new MeshBasicMaterial({
            vertexColors: FaceColors,
          });
          this.material.side = DoubleSide;
        }

        updatePoints(points, colors = [], options) {
          const vertices = [];
          const faces = [];
          const {
            scale: { x, y, z },
          } = options;

          this.scale.set(x, y, z);
          for (let index = 0, l = points.length / 3; index < l; index++) {
            const verticesArray = [
              points[3 * index],
              points[3 * index + 1],
              points[3 * index + 2],
            ];
            verticesArray.map(side => {
              vertices.push(new Vector3(side.x, side.y, side.z));
            });

            const color =
              colors.length === 0 ? { r: 1, g: 0, b: 0 } : colors[3 * index];
            faces.push(
              new Face3(
                3 * index,
                3 * index + 2,
                3 * index + 1,
                new Vector3(),
                new Color(color.r, color.g, color.b),
              ),
            );
          }

          this.geometry.vertices = vertices;
          this.geometry.faces = faces;

          this.geometry.computeFaceNormals();
          this.geometry.computeVertexNormals();
          this.geometry.elementsNeedUpdate = true;
          this.geometry.verticesNeedUpdate = true;
        }
      }

      class CubeList extends Mesh {
        constructor() {
          super();
          this.geometry = new Geometry();
          this.material = new MeshBasicMaterial();
          this.objectCacher = new ObjectCacher(this, Cube);
        }

        updatePoints(points, colors, options = {}) {
          options.subtype = MARKER_OBJECT_TYPES.CUBE;

          if (points.length < this.children.length) {
            this.objectCacher.reusePool(points, colors, options);
          } else {
            this.objectCacher.increasePool(points, colors, options);
          }
        }
      }

      class ViewFacingText extends Mesh {
        constructor(text) {
          super();
          this.color = new Color(1, 1, 1);
          this.text = new SpriteText(text, 1, `#${this.color.getHexString()}`);
          this.add(this.text);
        }

        setColor(color) {
          const { b, g, r } = color;
          if (!(this.color.r === r && this.color.g === g && this.color.b === b)) {
            this.color = new Color(r, g, b);
            this.text.color = `#${this.color.getHexString()}`;
          }
        }
      }

      const getNewPrimitive = marker => {
        switch (marker.type) {
          case MARKER_OBJECT_TYPES.CUBE:
            return new Cube();
          case MARKER_OBJECT_TYPES.SPHERE:
            return new Sphere();
          case MARKER_OBJECT_TYPES.CYLINDER: {
            const group = new Group();
            group.add(new Cylinder());
            return group;
          }
          case MARKER_OBJECT_TYPES.LINE_LIST:
            return new LineSegments();
          case MARKER_OBJECT_TYPES.LINE_STRIP:
            return new Line();
          case MARKER_OBJECT_TYPES.SPHERE_LIST:
            return new SphereList();
          case MARKER_OBJECT_TYPES.POINTS:
            return new Points();
          case MARKER_OBJECT_TYPES.TRIANGLE_LIST:
            return new TriangleList();
          case MARKER_OBJECT_TYPES.CUBE_LIST:
            return new CubeList();
          case MARKER_OBJECT_TYPES.TEXT_VIEW_FACING:
            return new ViewFacingText(marker.text);
          case MARKER_OBJECT_TYPES.ARROW:
          default: {
            const arrow = new Arrow();
            arrow.setHeadDimensions({
              radius: DEFAULT_CONE_RADIUS * 0.1,
              length: DEFAULT_CONE_HEIGHT * 0.3,
            });
            arrow.setShaftDimensions({
              radius: DEFAULT_CYLINDER_RADIUS * 0.05,
              length: DEFAULT_CYLINDER_HEIGHT * 0.7,
            });
            return arrow;
          }
        }
      };

      class MarkerLifetime {
        constructor(onTimeout) {
          this.timeouts = {};
          this.onTimeout = onTimeout;
        }

        track(id, timeoutSecs) {
          clearInterval(this.timeouts[id]);
          if (timeoutSecs === 0) {
            return;
          }
          this.timeouts[id] = setTimeout(() => {
            this.onTimeout(id);
          }, timeoutSecs * 1000);
        }

        untrack(id) {
          clearInterval(this.timeouts[id]);
          delete this.timeouts[id];
        }

        destroy() {
          const ids = Object.keys(this.timeouts);
          for (const id of ids) {
            clearInterval(this.timeouts[id]);
            delete this.timeouts[id];
          }
        }
      }

      class MarkerManager {
        constructor(rootObject, onChangeCb) {
          this.objectMap = {};
          this.object = rootObject;
          this.namespaces = {};
          this.onChangeCb = onChangeCb;
          this.markerLifetime = new MarkerLifetime(
            this.onMarkerLifetimeOver.bind(this),
          );
        }

        onMarkerLifetimeOver(id) {
          const marker = this.objectMap[id];
          if (!marker) {
            return;
          }

          this.removeObject(id);
        }

        getMarkerOrCreate(marker) {
          const id = MarkerManager.getId(marker);
          if (!this.objectMap[id]) {
            const object = getNewPrimitive(marker);
            this.objectMap[id] = object;
            this.object.add(object);
          }

          this.objectMap[id].visible = this.namespaces[marker.ns];
          return this.objectMap[id];
        }

        extractNameSpace(str) {
          const tokens = str.split('-');
          return tokens[0];
        }

        updateOptions(options) {
          const { namespaces } = options;
          this.namespaces = namespaces;
          for (const key in this.objectMap) {
            // eslint-disable-next-line no-prototype-builtins
            if (this.objectMap.hasOwnProperty(key)) {
              const namespace = this.extractNameSpace(key);
              this.objectMap[key].visible = this.namespaces[namespace];
            }
          }
        }

        onChange() {
          this.onChangeCb();
        }

        updateMarker(marker) {
          const {
            color,
            colors,
            lifetime,
            points,
            pose: { orientation, position },
            scale,
          } = marker;

          const markerObject = this.getMarkerOrCreate(marker);
          const markerId = MarkerManager.getId(marker);

          this.markerLifetime.track(markerId, lifetime.secs);

          if (markerObject.updatePoints) {
            markerObject.updatePoints(points, colors, marker);
          }

          markerObject.setTransform({ translation: position, rotation: orientation });

          // To avoid settings these properties for list types: LINE, TRIANGLE, CUBELIST etc
          if (markerObject.setScale && !markerObject.updatePoints) {
            markerObject.setScale({ x: scale.x, y: scale.y, z: scale.z });
          }
          if (markerObject.setColor && colors.length <= 0) {
            markerObject.setColor(color);
          }

          const { ns } = marker;
          if (!(ns in this.namespaces)) {
            this.namespaces[ns] = true;
            this.onChange();
          }
        }

        removeObject(id) {
          const obj = this.objectMap[id];
          obj.parent.remove(obj);
          delete this.objectMap[id];
        }

        reset() {
          this.namespaces = {};
          this.markerLifetime.destroy();
          this.onChange();

          Object.keys(this.objectMap).forEach(id => {
            this.removeObject(id);
          });
        }

        static getId({ ns, id }) {
          return `${ns}-${id}`;
        }
      }

      class MarkerArray extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_MARKERARRAY) {
          super(ros, topicName, MESSAGE_TYPE_MARKERARRAY, {
            ...DEFAULT_OPTIONS_MARKERARRAY,
            ...options,
          });

          this.object = new Group();
          this.onChange = this.onChange.bind(this);

          const { queueSize } = options;
          this.markerManager = new MarkerManager(this.object, this.onChange);
          this.queueSize = queueSize;
          this.updateOptions({
            ...DEFAULT_OPTIONS_MARKERARRAY,
            ...options,
          });
        }

        updateOptions(options) {
          super.updateOptions(options);
          this.markerManager.updateOptions(this.options, this);
        }

        onChange() {
          if (this.callback) {
            this.callback();
          }
        }

        onNamespaceChange(callback) {
          this.callback = callback;
        }

        update(message) {
          super.update(message);
          if (message.markers.length > 0) {
            message.markers.forEach(marker => {
              this.markerManager.updateMarker(marker);
            });
          }
        }

        reset() {
          this.markerManager.reset();
        }
      }

      class Points$1 {
        constructor(options = {}) {
          this.max_pts = options.max_pts || 10000;
          this.rootObject = options.rootObject || new Object3D();
        }

        setup(type, size, alpha) {
          this.rootObject.children.forEach(child => {
            this.rootObject.remove(child);
          });

          this.positions = new BufferAttribute(
            new Float32Array(this.max_pts * 3),
            3,
            false,
          );
          this.colors = new BufferAttribute(
            new Float32Array(this.max_pts * 3),
            3,
            false,
          );

          let options = {};

          if (type === LASERSCAN_STYLES.POINTS) {
            const sprite = new TextureLoader().load(
              'https://raw.githubusercontent.com/mrdoob/three.js/master/examples/textures/sprites/circle.png',
            );

            options = {
              map: sprite,
              alphaTest: 0.5,
            };
          }

          this.geomtry = new BufferGeometry();
          this.geomtry.addAttribute('position', this.positions.setDynamic(true));
          this.geomtry.addAttribute('color', this.colors.setDynamic(true));

          this.material = new PointsMaterial({
            color: 0x888888,
            size,
            ...options,
          });
          this.material.vertexColors = VertexColors$1;
          this.material.transparent = true;
          this.material.opacity = alpha;

          this.object = new Points$2(this.geomtry, this.material);
          this.rootObject.add(this.object);
        }

        update(data) {
          this.geomtry.setDrawRange(0, data);
          this.positions.needsUpdate = true;
          this.colors.needsUpdate = true;
          this.positions.updateRange.count = data * this.positions.itemSize;
          this.colors.updateRange.count = data * this.colors.itemSize;
        }

        setAlpha(alpha) {
          this.material.opacity = alpha;
        }

        setSize(size) {
          this.material.size = size;
        }
      }

      class LaserScan extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_LASERSCAN) {
          super(ros, topicName, MESSAGE_TYPE_LASERSCAN, options);

          this.points = new Points$1(LASERSCAN_STYLES.POINTS);
          this.sphereList = new SphereList();
          this.cubeList = new CubeList();

          this.object = new Group();
          this.object.add(this.points.rootObject);
          this.object.add(this.sphereList);
          this.object.add(this.cubeList);
          this.prevMessage = null;
          this.updateOptions({
            ...DEFAULT_OPTIONS_LASERSCAN,
            ...options,
          });
        }

        getNormalizedIntensity(data) {
          const { maxIntensity, minIntensity } = this.options;

          return (data - minIntensity) / (maxIntensity - minIntensity);
        }

        applyIntensityTransform(intensity, position) {
          const { channelName, maxColor, minColor } = this.options;
          const { x, y, z } = position;

          let normI;

          switch (channelName) {
            case INTENSITY_CHANNEL_OPTIONS.INTENSITY:
              normI = this.getNormalizedIntensity(intensity);
              break;
            case INTENSITY_CHANNEL_OPTIONS.X:
              normI = this.getNormalizedIntensity(x);
              break;
            case INTENSITY_CHANNEL_OPTIONS.Y:
              normI = this.getNormalizedIntensity(y);
              break;
            case INTENSITY_CHANNEL_OPTIONS.Z:
              normI = this.getNormalizedIntensity(z);
              break;
          }

          const minColorHex = new Color(minColor);
          const maxColorHex = new Color(maxColor);

          const finalColor =
            normI * maxColorHex.getHex() + (1 - normI) * minColorHex.getHex();
          return new Color(finalColor);
        }

        getNormalizedAxisValue(data) {
          const { maxAxisValue, minAxisValue } = this.options;

          return (data - minAxisValue) / (maxAxisValue - minAxisValue);
        }

        applyAxisColorTransform(intensity, position) {
          const { axis, maxAxisValue, minAxisValue } = this.options;
          const { x, y, z } = position;

          let normI;

          switch (axis) {
            case AXES.X:
              normI = this.getNormalizedAxisValue(x);
              break;
            case AXES.Y:
              normI = this.getNormalizedAxisValue(y);
              break;
            case AXES.Z:
              normI = this.getNormalizedAxisValue(z);
              break;
          }

          const finalColor = normI * maxAxisValue + (1 - normI) * minAxisValue;
          return new Color(finalColor);
        }

        colorTransformer(intensity, position) {
          const { colorTransformer, flatColor } = this.options;

          switch (colorTransformer) {
            case COLOR_TRANSFORMERS.INTENSITY:
              return this.applyIntensityTransform(intensity, position);
            case COLOR_TRANSFORMERS.AXIS_COLOR:
              return this.applyAxisColorTransform(intensity, position);
            case COLOR_TRANSFORMERS.FLAT_COLOR:
              return new Color(flatColor);
            default:
              return null;
          }
        }

        setupPoints({ j, position, color }) {
          this.points.colors.array[j] = color.r;
          this.points.positions.array[j++] = position.x;
          this.points.colors.array[j] = color.g;
          this.points.positions.array[j++] = position.y;
          this.points.colors.array[j] = color.b;
          this.points.positions.array[j++] = position.z;
        }

        hideAllObjects() {
          this.points.rootObject.visible = false;
          this.sphereList.visible = false;
          this.cubeList.visible = false;
        }

        setStyleDimensions(message) {
          if (!message) {
            return;
          }
          const { alpha, style } = this.options;
          const { size } = this.options;
          const { intensities, ranges } = message;
          const n = ranges.length;
          const positions = [];
          const colors = [];

          if (size < 0.001 || !size) {
            return;
          }

          this.hideAllObjects();
          this.points.setup(style, size, alpha);

          let j = 0;
          for (let i = 0; i < n; i++) {
            const range = message.ranges[i];

            if (range >= message.range_min && range <= message.range_max) {
              const angle = message.angle_min + i * message.angle_increment;
              const position = {
                x: range * Math.cos(angle),
                y: range * Math.sin(angle),
                z: 0,
              };
              const color = this.colorTransformer(intensities[i], position);

              switch (style) {
                case LASERSCAN_STYLES.POINTS:
                case LASERSCAN_STYLES.SQUARES:
                case LASERSCAN_STYLES.FLAT_SQUARES: {
                  this.setupPoints({ j, position, color });
                  j += 3;
                  break;
                }
                default:
                  positions.push(position);
                  colors.push(color);
                  break;
              }
            }
          }

          const options = { scale: { x: size, y: size, z: size } };

          switch (style) {
            case LASERSCAN_STYLES.SPHERES: {
              this.sphereList.visible = true;
              this.sphereList.updatePoints(positions, colors, options);
              break;
            }
            case LASERSCAN_STYLES.BOXES: {
              this.cubeList.visible = true;
              this.cubeList.updatePoints(positions, colors, options);
              break;
            }
            default:
              this.points.rootObject.visible = true;
              this.points.update(j / 3);
              break;
          }
        }

        updateOptions(options) {
          super.updateOptions(options);
          this.setStyleDimensions(this.prevMessage);
        }

        update(message) {
          super.update(message);
          this.setStyleDimensions(message);
          this.prevMessage = message;
        }
      }

      const populateImageDataFromNavMsg = (
        imageData,
        width,
        height,
        dataSource,
      ) => {
        for (let row = 0; row < height; row++) {
          for (let col = 0; col < width; col++) {
            const mapI = col + (height - row - 1) * width;
            const data = dataSource[mapI];
            const val = new DataView(Uint8Array.from([data]).buffer).getUint8(0);
            let i = (col + row * width) * 4;

            if (val >= 0 && val <= 100) {
              const v = 255 - (255 * val) / 100;
              imageData.data[i] = v; // red
              imageData.data[++i] = v; // green
              imageData.data[++i] = v; // blue
              imageData.data[++i] = 255; // alpha
            } else if (val >= 101 && val <= 127) {
              // illegal positive values in green
              imageData.data[i] = 0; // red
              imageData.data[++i] = 255; // green
              imageData.data[++i] = 0; // blue
              imageData.data[++i] = 255; // alpha
            } else if (val >= 128 && val <= 254) {
              // illegal negative (char) values in shades of red/yellow
              imageData.data[i] = 255; // red
              imageData.data[++i] = (255 * (val - 128)) / (254 - 128); // green
              imageData.data[++i] = 0; // blue
              imageData.data[++i] = 255; // alpha
            } else {
              // legal -1 value is tasteful blueish greenish grayish color
              imageData.data[i] = 0x70; // red
              imageData.data[++i] = 0x89; // green
              imageData.data[++i] = 0x86; // blue
              imageData.data[++i] = 255; // alpha
            }
          }
        }
      };

      const populateRawImageDataFromNavMsg = (
        imageData,
        width,
        height,
        dataSource,
      ) => {
        for (let row = 0; row < height; row++) {
          for (let col = 0; col < width; col++) {
            const mapI = col + (height - row - 1) * width;
            const data = dataSource[mapI];

            let i = (col + row * width) * 4;

            const Uint8DV = new DataView(Uint8Array.from([data]).buffer);
            const val = Uint8DV.getUint8(0);

            imageData.data[i] = val;
            imageData.data[++i] = val;
            imageData.data[++i] = val;
            imageData.data[++i] = 255;
          }
        }
      };

      const populateConstImageDataFromNavMsg = (
        imageData,
        width,
        height,
        dataSource,
      ) => {
        for (let row = 0; row < height; row++) {
          for (let col = 0; col < width; col++) {
            const mapI = col + (height - row - 1) * width;
            const data = dataSource[mapI];
            const val = new DataView(Uint8Array.from([data]).buffer).getUint8(0);

            let i = (col + row * width) * 4;

            if (val === 0) {
              imageData.data[i] = 0;
              imageData.data[++i] = 0;
              imageData.data[++i] = 0;
              imageData.data[++i] = 0; // alpha
            } else if (val >= 1 && val <= 98) {
              // Blue to red spectrum for most normal cost values
              const v = (255 * val) / 100;
              imageData.data[i] = v; // red
              imageData.data[++i] = 0; // green
              imageData.data[++i] = 255 - v; // blue
              imageData.data[++i] = 255; // alpha
            } else if (val === 99) {
              // inscribed obstacle values (99) in cyan
              imageData.data[i] = 0; // red
              imageData.data[++i] = 255; // green
              imageData.data[++i] = 255; // blue
              imageData.data[++i] = 255; // alpha
            } else if (val === 100) {
              // lethal obstacle values (100) in purple
              imageData.data[i] = 255; // red
              imageData.data[++i] = 0; // green
              imageData.data[++i] = 255; // blue
              imageData.data[++i] = 255; // alpha
            } else if (val > 100 && val <= 127) {
              // illegal positive values in green
              imageData.data[i] = 0; // red
              imageData.data[++i] = 255; // green
              imageData.data[++i] = 0; // blue
              imageData.data[++i] = 255; // alpha
            } else if (val >= 128 && val <= 254) {
              // illegal negative (char) values in shades of red/yellow
              imageData.data[i] = 255; // red
              imageData.data[++i] = (255 * (val - 128)) / (254 - 128); // green
              imageData.data[++i] = 0; // blue
              imageData.data[++i] = 255; // alpha
            } else {
              // legal -1 value is tasteful blueish greenish grayish color
              imageData.data[i] = 0x70; // red
              imageData.data[++i] = 0x89; // green
              imageData.data[++i] = 0x86; // blue
              imageData.data[++i] = 255; // alpha
            }
          }
        }
      };

      const imageDataToCanvas = imageData => {
        const canvas = document.createElement('canvas');
        canvas.width = imageData.width;
        canvas.height = imageData.height;
        const context = canvas.getContext('2d');
        context.putImageData(imageData, 0, 0);
        return canvas;
      };

      const populateImageDataFromImageMsg = (
        message,
        offset,
        rgbaOrder,
        imageData,
      ) => {
        const { data: rawData, height, step } = message;
        const typedArray = Uint8Array.from(rawData);
        // endianness is required for > 8bit encodings
        const encodedDataView = new DataView(typedArray.buffer);

        let j = 0;
        for (let i = 0; i < step * height; i += offset) {
          for (let k = 0; k < rgbaOrder.length; k++) {
            imageData.data[j++] = encodedDataView.getUint8(i + rgbaOrder[k]);
          }
          if (rgbaOrder.length === 3) {
            imageData.data[j++] = 255;
          }
        }
      };

      class Plane extends Mesh {
        constructor() {
          super();
          this.geometry = new PlaneGeometry();
          this.material = new MeshBasicMaterial();
        }
      }

      class Map$1 extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_MAP) {
          super(ros, topicName, MESSAGE_TYPE_OCCUPANCYGRID, {
            ...DEFAULT_OPTIONS_MAP,
            ...options,
          });
          this.object = new Plane();
          this.object.material.transparent = true;
          this.updateOptions({
            ...DEFAULT_OPTIONS_MAP,
            ...options,
          });
        }

        onMessage(callback) {
          this.callback = callback;
        }

        updateOptions(options) {
          super.updateOptions(options);
          const { alpha, drawBehind } = this.options;

          this.object.material.opacity = alpha;

          if (drawBehind) {
            this.object.material.side = DoubleSide;
          } else {
            this.object.material.side = FrontSide;
          }
          this.object.material.needsUpdate = true;
          if (this.prevMessage) {
            this.setCanvasData(this.prevMessage);
          }
        }

        updateCanvasDimensions(message) {
          const {
            info: { height, origin, resolution, width },
          } = message;

          this.object.scale.set(width * resolution, -1 * height * resolution, 1);
          const translatedX = (width * resolution) / 2 + origin.position.x;
          const translatedY = (height * resolution) / 2 + origin.position.y;
          this.object.position.set(translatedX, translatedY, 0);
        }

        setCanvasData(message) {
          const { colorScheme } = this.options;
          const {
            data,
            info: { height, width },
          } = message;

          const imageData = new ImageData(width, height);
          let bitmapCanvas = null;

          switch (colorScheme) {
            case MAP_COLOR_SCHEMES.MAP:
              populateImageDataFromNavMsg(imageData, width, height, data);
              bitmapCanvas = imageDataToCanvas(imageData);
              break;
            case MAP_COLOR_SCHEMES.CONST_MAP:
              populateConstImageDataFromNavMsg(imageData, width, height, data);
              bitmapCanvas = imageDataToCanvas(imageData);
              break;
            case MAP_COLOR_SCHEMES.RAW:
              populateRawImageDataFromNavMsg(imageData, width, height, data);
              bitmapCanvas = imageDataToCanvas(imageData);
              break;
          }

          this.object.material.map = new CanvasTexture(bitmapCanvas);
          this.object.material.map.minFilter = NearestFilter;
          this.object.material.map.magFilter = NearestFilter;
          this.object.material.needsUpdate = true;

          this.updateCanvasDimensions(message);
        }

        update(message) {
          super.update(message);
          if (this.callback) {
            this.callback(message);
          }

          this.setCanvasData(message);
          this.prevMessage = message;
        }
      }

      class Odometry extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_ODOMETRY) {
          super(ros, topicName, MESSAGE_TYPE_ODOMETRY, {
            ...DEFAULT_OPTIONS_ODOMETRY,
            ...options,
          });

          this.object = null;
          this.objectPool = [];
          this.keepSize = 100;
          this.currentObject = -1;
          this.setVizType(options.controlledObject);
          this.updateOptions({
            ...DEFAULT_OPTIONS_ODOMETRY,
            ...options,
          });
        }

        setVizType(controlledObject) {
          if (controlledObject) {
            this.object = controlledObject;
          } else {
            this.object = new Group();
          }
        }

        setKeepSize(size) {
          let newKeepList = [];

          if (size === 0) {
            this.keepSize = 0;
            return;
          }

          if (size < this.keepSize && size < this.objectPool.length) {
            const removeCount = this.objectPool.length - size;
            for (let i = 0; i < removeCount; i++) {
              this.object.remove(this.objectPool[i]);
            }

            const slicedList = this.objectPool.slice(
              this.objectPool.length - size,
              this.objectPool.length,
            );
            newKeepList = [...slicedList];
          } else {
            newKeepList = [...this.objectPool];
          }

          this.objectPool = newKeepList;
          this.keepSize = size;
          this.currentObject = this.objectPool.length - 1;
        }

        removeAllObjects() {
          this.objectPool.forEach((obj, index) => {
            obj.parent.remove(obj);
            delete this.objectPool[index];
          });
          this.objectPool = [];
        }

        checkToleranceThresholdExceed(newPose) {
          if (this.objectPool.length === 0) {
            return true;
          }

          const oldPose = {
            position: this.objectPool[this.currentObject].position,
            quaternion: this.objectPool[this.currentObject].quaternion,
          };

          return checkToleranceThresholdExceed(oldPose, newPose, this.options);
        }

        getObject() {
          const { type } = this.options;
          switch (type) {
            case ODOMETRY_OBJECT_TYPES.arrow:
              return new Arrow();
            case ODOMETRY_OBJECT_TYPES.axes:
              return new Axes();
          }

          return new Object3D();
        }

        changeObjectPoolType() {
          const tempObjectPool = [];

          // remove prev type objects and push the new ones in place of them.
          this.objectPool.forEach((object, index) => {
            const { position, quaternion } = object;
            object.parent.remove(object);
            delete this.objectPool[index];

            const newObj = this.getObject();
            newObj.position.set(position.x, position.y, position.z);
            newObj.quaternion.set(
              quaternion.x,
              quaternion.y,
              quaternion.z,
              quaternion.w,
            );
            tempObjectPool.push(newObj);
            this.object.add(newObj);
            setObjectDimension(newObj, this.options);
          });

          this.objectPool = tempObjectPool;
        }

        updateOptions(options) {
          const { type: currentType } = this.options;
          super.updateOptions(options);
          const { keep, type } = this.options;

          if (type !== currentType) {
            this.changeObjectPoolType();
          }

          this.objectPool.forEach(object => {
            setObjectDimension(object, this.options);
          });

          this.setKeepSize(keep);
        }

        update(message) {
          super.update(message);
          if (!this.keepSize) {
            this.removeAllObjects();
            return;
          }

          const {
            pose: {
              pose: { orientation, position },
            },
          } = message;
          const transform = {
            translation: position,
            rotation: orientation,
          };

          const newPose = {
            position: new Vector3(position.x, position.y, position.z),
            quaternion: new Quaternion(
              orientation.x,
              orientation.y,
              orientation.z,
              orientation.w,
            ),
          };
          const toleranceThresholdExceed = this.checkToleranceThresholdExceed(
            newPose,
          );

          if (toleranceThresholdExceed) {
            const newObject = this.getObject();
            setObjectDimension(newObject, this.options);

            this.objectPool.push(newObject);
            this.currentObject += 1;
            this.currentObject = Math$1.clamp(this.currentObject, 0, this.keepSize - 1);
            this.object.add(newObject);
            setTransform(newObject, transform);

            // remove excess object from object pool wrt to keepsize
            if (this.objectPool.length > this.keepSize) {
              const objToRemove = this.objectPool[0];
              this.object.remove(objToRemove);
              delete this.objectPool[0];

              const newObjectPool = this.objectPool.slice(1);
              this.objectPool = [...newObjectPool];
            }
          }
        }
      }

      class PoseArray extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_POSEARRAY) {
          super(ros, topicName, MESSAGE_TYPE_POSEARRAY, {
            ...DEFAULT_OPTIONS_POSEARRAY,
            ...options,
          });
          this.object = new Group$1();
          this.updateOptions({
            ...DEFAULT_OPTIONS_POSEARRAY,
            ...options,
          });
        }

        update(message) {
          super.update(message);
          this.object.children.forEach(obj => {
            obj.parent.remove(obj);
          });
          this.object.children = [];

          for (let i = 0; i < message.poses.length; i++) {
            this.object.add(Pose.getNewPrimitive(this.options));
          }

          for (let i = 0; i < message.poses.length; i++) {
            setTransform(this.object.children[i], {
              translation: message.poses[i].position,
              rotation: message.poses[i].orientation,
            });
            setObjectDimension(this.object.children[i], this.options);
          }
        }
      }

      class Path extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_PATH) {
          super(ros, topicName, MESSAGE_TYPE_PATH, {
            ...DEFAULT_OPTIONS_PATH,
            ...options,
          });
          this.object = new Group();
          this.line = new Line(null, true);
          this.updateOptions({
            ...DEFAULT_OPTIONS_PATH,
            ...options,
          });
        }

        updateOptions(options) {
          super.updateOptions(options);
          const { alpha, color } = this.options;
          this.line.setColor(new Color(color));
          this.line.setAlpha(alpha);
        }

        update(message) {
          super.update(message);
          const { poses } = message;
          const points = (poses || []).map(poseData => poseData.pose.position);

          this.line.updatePoints(points);
          this.object.add(this.line);
        }
      }

      class Image extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_IMAGE) {
          super(ros, topicName, MESSAGE_TYPE_IMAGE, {
            ...DEFAULT_OPTIONS_IMAGE,
            ...options,
          });
          this.object = document.createElement('canvas');
          this.shadowObject = document.createElement('canvas');
          this.updateOptions({
            ...DEFAULT_OPTIONS_IMAGE,
            ...options,
          });
        }

        applyImageData(message) {
          const { encoding, height, width } = message;

          const aspectRatio = width / height;
          const ctx = this.object.getContext('2d');
          const shadowCtx = this.shadowObject.getContext('2d');
          const imgData = shadowCtx.createImageData(width, height);

          switch (encoding) {
            // not using encoding.find(bayer) because js switch statement
            // is much faster
            case 'bayer_rggb8':
            case 'bayer_bggr8':
            case 'bayer_gbrg8':
            case 'bayer_grbg8':
            case 'bayer_rggb16':
            case 'bayer_bggr16':
            case 'bayer_gbrg16':
            case 'bayer_grbg16':
            case '8UC1':
            case '8SC1':
            case 'mono8': {
              populateImageDataFromImageMsg(message, 1, [0, 0, 0], imgData);
              break;
            }
            case '8UC3':
            case '8SC3':
            case 'bgr8': {
              populateImageDataFromImageMsg(message, 3, [2, 1, 0], imgData);
              break;
            }
            case '8UC4':
            case '8SC4':
            case 'bgra8': {
              populateImageDataFromImageMsg(message, 4, [2, 1, 0, 3], imgData);
              break;
            }
            case 'rgb8': {
              populateImageDataFromImageMsg(message, 3, [0, 1, 2], imgData);
              break;
            }
            case 'rgba8': {
              populateImageDataFromImageMsg(message, 4, [0, 1, 2, 3], imgData);
              break;
            }
          }

          ctx.clearRect(0, 0, this.object.width, this.object.height);
          ctx.fillStyle = '#000';
          ctx.fillRect(0, 0, this.object.width, this.object.height);
          shadowCtx.clearRect(
            0,
            0,
            this.shadowObject.width,
            this.shadowObject.height,
          );
          shadowCtx.putImageData(imgData, 0, 0);

          const scaledImageHeight = this.object.width / aspectRatio;
          const scaledImageWidth = this.object.height * aspectRatio;

          if (aspectRatio >= this.object.width / this.object.height) {
            ctx.drawImage(
              this.shadowObject,
              0,
              0,
              width,
              height,
              0,
              (this.object.height - scaledImageHeight) / 2,
              this.object.width,
              scaledImageHeight,
            );
          } else {
            ctx.drawImage(
              this.shadowObject,
              0,
              0,
              width,
              height,
              (this.object.width - scaledImageWidth) / 2,
              0,
              scaledImageWidth,
              this.object.height,
            );
          }
        }

        updateDimensions(width, height) {
          this.object.width = width;
          this.object.height = height;
        }

        updateOptions(options) {
          super.updateOptions(options);
          const { height, width } = this.options;
          this.object.width = width;
          this.object.height = height;
        }

        update(message) {
          const { height, width } = message;
          this.shadowObject.width = width;
          this.shadowObject.height = height;
          this.applyImageData(message);
        }

        hide() {
          this.object.style.visibility = 'hidden';
        }

        show() {
          this.object.style.visibility = 'visible';
        }
      }

      class ImageStream extends Core {
        constructor(url, options = DEFAULT_OPTIONS_IMAGE_STREAM) {
          super(null, null, null, {
            ...DEFAULT_OPTIONS_IMAGE_STREAM,
            ...options,
          });
          this.url = url;
          this.object = document.createElement('img');
          this.object.src = this.url;
          this.object.setAttribute('draggable', false);
          this.updateOptions({
            ...DEFAULT_OPTIONS_IMAGE_STREAM,
            ...options,
          });
        }

        updateOptions(options) {
          super.updateOptions(options);
          const { height, width } = this.options;
          this.object.width = width;
          this.object.height = height;
        }

        hide() {
          this.object.style.visibility = 'hidden';
        }

        show() {
          this.object.style.visibility = 'visible';
        }
      }

      class Marker extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_MARKER) {
          super(ros, topicName, MESSAGE_TYPE_MARKER, {
            ...DEFAULT_OPTIONS_MARKER,
            ...options,
          });

          this.object = new Group();
          this.onChange = this.onChange.bind(this);

          const { queueSize } = options;
          this.markerManager = new MarkerManager(this.object, this.onChange);
          this.queueSize = queueSize;
          this.updateOptions({
            ...DEFAULT_OPTIONS_MARKER,
            ...options,
          });
        }

        updateOptions(options) {
          super.updateOptions(options);
          this.markerManager.updateOptions(this.options, this);
        }

        update(message) {
          super.update(message);
          this.markerManager.updateMarker(message);
        }

        onChange() {
          if (this.callback) {
            this.callback();
          }
        }

        onNamespaceChange(callback) {
          this.callback = callback;
        }

        reset() {
          this.markerManager.reset();
        }
      }

      class Range extends Core {
        constructor(ros, topicName, options = DEFAULT_OPTIONS_RANGE) {
          super(ros, topicName, MESSAGE_TYPE_RANGE, options);
          this.object = new Group();
          this.cylinder = new Cylinder(this.options.color, 0.01, 0.01);
          this.cylinder.rotateZ(Math.PI / 2);
          this.object.add(this.cylinder);
          this.updateOptions({
            ...DEFAULT_OPTIONS_RANGE,
            ...options,
          });
        }

        updateOptions(options) {
          super.updateOptions(options);
          const { alpha, color } = this.options;
          this.cylinder.setAlpha(alpha);
          this.cylinder.setColor(new Color(color));
        }

        update(message) {
          super.update(message);
          const { field_of_view: fov, max_range: max, min_range: min } = message;
          this.cylinder.geometry = new CylinderGeometry(
            (min * Math.tan(fov)) / 4,
            (max * Math.tan(fov)) / 4,
            max - min,
            DEFAULT_RADIAL_SEGMENTS,
            DEFAULT_RADIAL_SEGMENTS,
          );
          this.cylinder.geometryNeedsUpdate = true;
          this.cylinder.position.setX(0.5 * (min + max));
        }
      }

      var stats_min = createCommonjsModule(function (module, exports) {
      !function(e,t){module.exports=t();}(commonjsGlobal,function(){var c=function(){var n=0,l=document.createElement("div");function e(e){return l.appendChild(e.dom),e}function t(e){for(var t=0;t<l.children.length;t++)l.children[t].style.display=t===e?"block":"none";n=e;}l.style.cssText="position:fixed;top:0;left:0;cursor:pointer;opacity:0.9;z-index:10000",l.addEventListener("click",function(e){e.preventDefault(),t(++n%l.children.length);},!1);var i=(performance||Date).now(),a=i,o=0,f=e(new c.Panel("FPS","#0ff","#002")),r=e(new c.Panel("MS","#0f0","#020"));if(self.performance&&self.performance.memory)var d=e(new c.Panel("MB","#f08","#201"));return t(0),{REVISION:16,dom:l,addPanel:e,showPanel:t,begin:function(){i=(performance||Date).now();},end:function(){o++;var e=(performance||Date).now();if(r.update(e-i,200),a+1e3<=e&&(f.update(1e3*o/(e-a),100),a=e,o=0,d)){var t=performance.memory;d.update(t.usedJSHeapSize/1048576,t.jsHeapSizeLimit/1048576);}return e},update:function(){i=this.end();},domElement:l,setMode:t}};return c.Panel=function(n,l,i){var a=1/0,o=0,f=Math.round,r=f(window.devicePixelRatio||1),d=80*r,e=48*r,c=3*r,p=2*r,u=3*r,s=15*r,m=74*r,h=30*r,y=document.createElement("canvas");y.width=d,y.height=e,y.style.cssText="width:80px;height:48px";var v=y.getContext("2d");return v.font="bold "+9*r+"px Helvetica,Arial,sans-serif",v.textBaseline="top",v.fillStyle=i,v.fillRect(0,0,d,e),v.fillStyle=l,v.fillText(n,c,p),v.fillRect(u,s,m,h),v.fillStyle=i,v.globalAlpha=.9,v.fillRect(u,s,m,h),{dom:y,update:function(e,t){a=Math.min(a,e),o=Math.max(o,e),v.fillStyle=i,v.globalAlpha=1,v.fillRect(0,0,d,s),v.fillStyle=l,v.fillText(f(e)+" "+n+" ("+f(a)+"-"+f(o)+")",c,p),v.drawImage(y,u+r,s,m-r,h,u,s,m-r,h),v.fillRect(u+m-r,s,r,h),v.fillStyle=i,v.globalAlpha=.9,v.fillRect(u+m-r,s,r,f((1-e/t)*h));}}},c});
      });

      class Scene extends Scene$1 {
        constructor(options = {}) {
          super();
          this.vizWrapper = new Group$1();
          this.add(this.vizWrapper);

          this.stats = new stats_min();
          this.stats.showPanel(0);

          this.initLights();
          this.initGrid();
          this.updateOptions(options);
        }

        initLights() {
          [
            [-1, 0],
            [1, 0],
            [0, -1],
            [0, 1],
          ].forEach(positions => {
            const directionalLight = new DirectionalLight(0xffffff, 0.4);
            [directionalLight.position.x, directionalLight.position.y] = positions;
            directionalLight.position.z = 1;
            this.add(directionalLight);
          });
          const ambientLight = new AmbientLight(0xffffff, 0.2);
          this.add(ambientLight);
        }

        initGrid() {
          this.grid = new GridHelper(0, 0);
          this.add(this.grid);
        }

        addObject(object) {
          this.vizWrapper.add(object);
        }

        addVisualization({ object }) {
          if (!this.vizWrapper.getObjectById(object.id)) {
            this.addObject(object);
          }
        }

        getObjectByName(name) {
          return this.vizWrapper.getObjectByName(name);
        }

        updateOptions(options) {
          this.options = {
            ...DEFAULT_OPTIONS_SCENE,
            ...options,
          };

          const {
            backgroundColor,
            gridCenterlineColor,
            gridColor,
            gridDivisions,
            gridSize,
          } = this.options;

          this.grid.copy(
            new GridHelper(gridSize, gridDivisions, gridCenterlineColor, gridColor),
          );
          this.grid.rotateX(-Math.PI / 2);

          this.background = new Color(backgroundColor);
        }
      }

      /**
       * Software License Agreement (BSD License)

       Copyright (c) 2014, Worcester Polytechnic Institute, Robert Bosch
       LLC, Yujin Robot. All rights reserved.

       Redistribution and use in source and binary forms, with or without
       modification, are permitted provided that the following conditions
       are met:

       * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
       * Redistributions in binary form must reproduce the above
       copyright notice, this list of conditions and the following
       disclaimer in the documentation and/or other materials provided
       with the distribution.
       * Neither the name of Worcester Polytechnic Institute, Robert
       Bosch LLC, Yujin Robot nor the names of its contributors may be
       used to endorse or promote products derived from this software
       without specific prior written permission.

       THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
       "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
       LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
       FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
       COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
       INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
       BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
       LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
       CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
       LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
       ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
       POSSIBILITY OF SUCH DAMAGE.
       */

      const vertexShader = `uniform sampler2D map;

uniform float width;
uniform float height;
uniform float nearClipping, farClipping;

uniform float pointSize;
uniform float zOffset;

uniform float focallength;
uniform float maxDepthPerTile;
uniform float resolutionFactor;

varying vec2 vUvP;
varying vec2 colorP;

varying float depthVariance;
varying float maskVal;

float sampleDepth(vec2 pos)
  {
    float depth;
    
    vec2 vUv = vec2( pos.x / (width*2.0), pos.y / (height*2.0)+0.5 );
    vec2 vUv2 = vec2( pos.x / (width*2.0)+0.5, pos.y / (height*2.0)+0.5 );
    
    vec4 depthColor = texture2D( map, vUv );
    
    depth = ( depthColor.r + depthColor.g + depthColor.b ) / 3.0 ;
    
    if (depth>0.99)
    {
      vec4 depthColor2 = texture2D( map, vUv2 );
      float depth2 = ( depthColor2.r + depthColor2.g + depthColor2.b ) / 3.0 ;
      depth = 0.99+depth2;
    }
    
    return depth;
  }

float median(float a, float b, float c)
  {
    float r=a;
    
    if ( (a<b) && (b<c) )
    {
      r = b;
    }
    if ( (a<c) && (c<b) )
    {
      r = c;
    }
    return r;
  }

float variance(float d1, float d2, float d3, float d4, float d5, float d6, float d7, float d8, float d9)
  {
    float mean = (d1 + d2 + d3 + d4 + d5 + d6 + d7 + d8 + d9) / 9.0;
    float t1 = (d1-mean);
    float t2 = (d2-mean);
    float t3 = (d3-mean);
    float t4 = (d4-mean);
    float t5 = (d5-mean);
    float t6 = (d6-mean);
    float t7 = (d7-mean);
    float t8 = (d8-mean);
    float t9 = (d9-mean);
    float v = (t1*t1+t2*t2+t3*t3+t4*t4+t5*t5+t6*t6+t7*t7+t8*t8+t9*t9)/9.0;
    return v;
  }

vec2 decodeDepth(vec2 pos)
  {
    vec2 ret;
    
    
    float depth1 = sampleDepth(vec2(position.x-1.0, position.y-1.0));
    float depth2 = sampleDepth(vec2(position.x, position.y-1.0));
    float depth3 = sampleDepth(vec2(position.x+1.0, position.y-1.0));
    float depth4 = sampleDepth(vec2(position.x-1.0, position.y));
    float depth5 = sampleDepth(vec2(position.x, position.y));
    float depth6 = sampleDepth(vec2(position.x+1.0, position.y));
    float depth7 = sampleDepth(vec2(position.x-1.0, position.y+1.0));
    float depth8 = sampleDepth(vec2(position.x, position.y+1.0));
    float depth9 = sampleDepth(vec2(position.x+1.0, position.y+1.0));
    
    float median1 = median(depth1, depth2, depth3);
    float median2 = median(depth4, depth5, depth6);
    float median3 = median(depth7, depth8, depth9);
    
    ret.x = median(median1, median2, median3);
    ret.y = variance(depth1, depth2, depth3, depth4, depth5, depth6, depth7, depth8, depth9);
    
    return ret;
    
  }


void main() {
  
  vUvP = vec2( position.x / (width*2.0), position.y / (height*2.0)+0.5 );
  colorP = vec2( position.x / (width*2.0)+0.5 , position.y / (height*2.0)  );
  
  vec4 pos = vec4(0.0,0.0,0.0,0.0);
  depthVariance = 0.0;
  
  if ( (vUvP.x<0.0)|| (vUvP.x>0.5) || (vUvP.y<0.5) || (vUvP.y>0.0))
  {
    vec2 smp = decodeDepth(vec2(position.x, position.y));
    float depth = smp.x;
    depthVariance = smp.y;
    
    float z = -depth;
    
    pos = vec4(
      ( position.x / width - 0.5 ) * z * 0.5 * maxDepthPerTile * resolutionFactor * (1000.0/focallength) * -1.0,
      ( position.y / height - 0.5 ) * z * 0.5 * maxDepthPerTile * resolutionFactor * (1000.0/focallength),
      (- z + zOffset / 1000.0) * maxDepthPerTile,
      1.0);
    
    vec2 maskP = vec2( position.x / (width*2.0), position.y / (height*2.0)  );
    vec4 maskColor = texture2D( map, maskP );
    maskVal = ( maskColor.r + maskColor.g + maskColor.b ) / 3.0 ;
  }
  
  gl_PointSize = pointSize;
  gl_Position = projectionMatrix * modelViewMatrix * pos;
  
}`;

      const fragmentShader = `uniform sampler2D map;
uniform float varianceThreshold;
uniform float whiteness;

varying vec2 vUvP;
varying vec2 colorP;

varying float depthVariance;
varying float maskVal;


void main() {
  
  vec4 color;
  
  if ( (depthVariance>varianceThreshold) || (maskVal>0.5) ||(vUvP.x<0.0)|| (vUvP.x>0.5) || (vUvP.y<0.5) || (vUvP.y>1.0))
  {  
    discard;
  }
  else 
  {
    color = texture2D( map, colorP );
    
    float fader = whiteness /100.0;
    
    color.r = color.r * (1.0-fader)+ fader;
    
    color.g = color.g * (1.0-fader)+ fader;
    
    color.b = color.b * (1.0-fader)+ fader;
    
    color.a = 1.0;//smoothstep( 20000.0, -20000.0, gl_FragCoord.z / gl_FragCoord.w );
  }
  
  gl_FragColor = vec4( color.r, color.g, color.b, color.a );
  
}`;

      class DepthCloudObject extends Object3D {
        constructor(url, options = DEFAULT_OPTIONS_DEPTHCLOUD) {
          super();
          this.options = {
            ...DEFAULT_OPTIONS_DEPTHCLOUD,
            ...options,
          };
          this.url = url;

          this.animate = this.animate.bind(this);
          this.onMetaLoaded = this.onMetaLoaded.bind(this);
        }

        onMetaLoaded() {
          this.metaLoaded = true;
          this.initStreamer();
        }

        isMjpeg() {
          const { streamType } = this.options;
          return streamType.toLowerCase() === 'mjpeg';
        }

        initVideo() {
          const { height, width } = this.options;

          this.video = document.createElement(this.isMjpeg() ? 'img' : 'video');
          this.video.width = width;
          this.video.height = height;
          this.video.addEventListener(
            this.isMjpeg() ? 'load' : 'loadedmetadata',
            this.onMetaLoaded,
            false,
          );

          if (!this.isMjpeg()) {
            this.video.loop = true;
          }

          this.video.src = this.url;
          this.video.crossOrigin = 'Anonymous';
          this.video.setAttribute('crossorigin', 'Anonymous');
        }

        initStreamer() {
          const {
            f,
            height,
            maxDepthPerTile,
            pointSize,
            varianceThreshold,
            whiteness,
            width,
          } = this.options;
          const resolutionFactor = Math.max(width, height) / 1024;
          if (this.metaLoaded) {
            this.texture = new Texture(this.video);
            this.geometry = new Geometry();

            const halfWidth = width / 2;
            const halfHeight = height / 2;
            for (let i = 0, l = halfWidth * halfHeight; i < l; i += 1) {
              this.geometry.vertices.push(
                new Vector3(i % halfWidth, i / halfWidth, 0),
              );
            }

            this.material = new ShaderMaterial({
              uniforms: {
                map: { type: 't', value: this.texture },
                width: { type: 'f', value: halfWidth },
                height: { type: 'f', value: halfHeight },
                focallength: { type: 'f', value: f },
                pointSize: { type: 'f', value: pointSize },
                zOffset: { type: 'f', value: 0 },
                whiteness: { type: 'f', value: whiteness },
                varianceThreshold: { type: 'f', value: varianceThreshold },
                maxDepthPerTile: { type: 'f', value: maxDepthPerTile },
                resolutionFactor: { type: 'f', value: resolutionFactor },
              },
              vertexShader,
              fragmentShader,
            });

            this.mesh = new Points$2(this.geometry, this.material);
            this.mesh.frustumCulled = false;
            this.mesh.position.set(0, 0, 0);

            this.add(this.mesh);

            requestAnimationFrame(this.animate);
          }
        }

        animate() {
          if (
            this.isMjpeg() ||
            this.video.readyState === this.video.HAVE_ENOUGH_DATA
          ) {
            this.texture.needsUpdate = true;
          }
          if (this.video.src) {
            requestAnimationFrame(this.animate);
          }
        }

        startStream() {
          if (!this.isMjpeg()) {
            this.video.play();
          }
        }

        stopStream() {
          if (!this.isMjpeg()) {
            this.video.pause();
          }
          this.video.removeAttribute('src');
        }
      }

      class DepthCloud extends Core {
        constructor(url, options) {
          super(null, url, null, {
            ...DEFAULT_OPTIONS_DEPTHCLOUD,
            ...options,
          });
          this.url = url;
          this.object = new DepthCloudObject(url, options);
        }

        subscribe() {
          this.object.initVideo();
          this.object.startStream();
        }

        unsubscribe() {
          this.object.stopStream();
        }
      }

      // @ts-nocheck
      var MapControls2D = function (object, domElement) {
          domElement = domElement !== undefined ? domElement : document;
          // API
          this.enabled = true;
          this.center = new Vector3();
          this.panSpeed = 0.01;
          this.zoomSpeed = 0.05;
          this.rotationSpeed = 0.005;
          // internals
          var scope = this;
          var vector = new Vector3();
          var delta = new Vector3();
          var box = new Box3();
          var STATE = { NONE: -1, ROTATE: 0, ZOOM: 1, PAN: 2 };
          var state = STATE.NONE;
          var center = this.center;
          var normalMatrix = new Matrix3();
          var pointer = new Vector2();
          var pointerOld = new Vector2();
          var spherical = new Spherical();
          var sphere = new Sphere$1();
          // events
          var changeEvent = { type: 'change' };
          this.focus = function (target) {
              var distance;
              box.setFromObject(target);
              if (box.isEmpty() === false) {
                  box.getCenter(center);
                  distance = box.getBoundingSphere(sphere).radius;
              }
              else {
                  // Focusing on an Group, AmbientLight, etc
                  center.setFromMatrixPosition(target.matrixWorld);
                  distance = 0.1;
              }
              delta.set(0, 0, 1);
              delta.applyQuaternion(object.quaternion);
              delta.multiplyScalar(distance * 4);
              object.position.copy(center).add(delta);
              scope.dispatchEvent(changeEvent);
          };
          this.pan = function (positionDelta) {
              // var distance = object.position.distanceTo( center );
              var zoom = object.zoom;
              positionDelta.multiplyScalar(scope.panSpeed / zoom);
              positionDelta.applyMatrix3(normalMatrix.getNormalMatrix(object.matrix));
              object.position.add(positionDelta);
              center.add(positionDelta);
              scope.dispatchEvent(changeEvent);
          };
          this.zoom = function (zoomDelta) {
              object.zoom -= (zoomDelta.z || 0) * object.zoom * 0.1;
              object.updateProjectionMatrix();
              scope.dispatchEvent(changeEvent);
          };
          this.rotate = function (rotationDelta) {
              vector.copy(object.position).sub(center);
              // spherical.setFromVector3( vector );
              spherical.setFromCartesianCoords(-1 * vector.x, vector.z, vector.y);
              spherical.theta += rotationDelta.x * scope.rotationSpeed;
              // spherical.phi += rotationDelta.y * scope.rotationSpeed;
              spherical.makeSafe();
              var tempRelPosition = vector.setFromSpherical(spherical);
              vector.set(-1 * tempRelPosition.x, tempRelPosition.z, tempRelPosition.y);
              object.position.copy(center).add(vector);
              object.lookAt(center);
              scope.dispatchEvent(changeEvent);
          };
          // mouse
          function onMouseMove(event) {
              if (scope.enabled === false)
                  return;
              pointer.set(event.clientX, event.clientY);
              var movementX = pointer.x - pointerOld.x;
              var movementY = pointer.y - pointerOld.y;
              if (state === STATE.ROTATE) {
                  scope.rotate(delta.set(-movementX, -movementY, 0));
              }
              else if (state === STATE.ZOOM) {
                  scope.zoom(delta.set(0, 0, movementY));
              }
              else if (state === STATE.PAN) {
                  scope.pan(delta.set(-movementX, movementY, 0));
              }
              pointerOld.set(event.clientX, event.clientY);
          }
          function onMouseUp() {
              domElement.removeEventListener('mousemove', onMouseMove, false);
              domElement.removeEventListener('mouseup', onMouseUp, false);
              domElement.removeEventListener('mouseout', onMouseUp, false);
              domElement.removeEventListener('dblclick', onMouseUp, false);
              state = STATE.NONE;
          }
          function onMouseWheel(event) {
              event.preventDefault();
              // Normalize deltaY due to https://bugzilla.mozilla.org/show_bug.cgi?id=1392460
              scope.zoom(delta.set(0, 0, event.deltaY > 0 ? 1 : -1));
          }
          function onMouseDown(event) {
              if (scope.enabled === false)
                  return;
              if (event.button === 0) {
                  state = STATE.ROTATE;
              }
              else if (event.button === 1) {
                  state = STATE.ZOOM;
              }
              else if (event.button === 2) {
                  state = STATE.PAN;
              }
              pointerOld.set(event.clientX, event.clientY);
              domElement.addEventListener('mousemove', onMouseMove, false);
              domElement.addEventListener('mouseup', onMouseUp, false);
              domElement.addEventListener('mouseout', onMouseUp, false);
              domElement.addEventListener('dblclick', onMouseUp, false);
          }
          function contextmenu(event) {
              event.preventDefault();
          }
          domElement.addEventListener('contextmenu', contextmenu, false);
          domElement.addEventListener('mousedown', onMouseDown, false);
          domElement.addEventListener('wheel', onMouseWheel, false);
          // touch
          var touches = [new Vector3(), new Vector3(), new Vector3()];
          var prevTouches = [new Vector3(), new Vector3(), new Vector3()];
          var prevDistance = null;
          function touchStart(event) {
              if (scope.enabled === false)
                  return;
              switch (event.touches.length) {
                  case 1:
                      touches[0]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      touches[1]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      break;
                  case 2:
                      touches[0]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      touches[1]
                          .set(event.touches[1].pageX, event.touches[1].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      prevDistance = touches[0].distanceTo(touches[1]);
                      break;
              }
              prevTouches[0].copy(touches[0]);
              prevTouches[1].copy(touches[1]);
          }
          function touchMove(event) {
              if (scope.enabled === false)
                  return;
              event.preventDefault();
              event.stopPropagation();
              function getClosest(touch, allTouches) {
                  var closest = allTouches[0];
                  for (var i in allTouches) {
                      if (closest.distanceTo(touch) > allTouches[i].distanceTo(touch)) {
                          closest = allTouches[i];
                      }
                  }
                  return closest;
              }
              switch (event.touches.length) {
                  case 1:
                      touches[0]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      touches[1]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      scope.rotate(touches[0]
                          .sub(getClosest(touches[0], prevTouches))
                          .multiplyScalar(-1));
                      break;
                  case 2:
                      touches[0]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      touches[1]
                          .set(event.touches[1].pageX, event.touches[1].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      var distance = touches[0].distanceTo(touches[1]);
                      scope.zoom(delta.set(0, 0, prevDistance - distance));
                      prevDistance = distance;
                      var offset0 = touches[0]
                          .clone()
                          .sub(getClosest(touches[0], prevTouches));
                      var offset1 = touches[1]
                          .clone()
                          .sub(getClosest(touches[1], prevTouches));
                      offset0.x = -offset0.x;
                      offset1.x = -offset1.x;
                      scope.pan(offset0.add(offset1));
                      break;
              }
              prevTouches[0].copy(touches[0]);
              prevTouches[1].copy(touches[1]);
          }
          domElement.addEventListener('touchstart', touchStart, false);
          domElement.addEventListener('touchmove', touchMove, false);
          this.dispose = function () {
              domElement.removeEventListener('contextmenu', contextmenu, false);
              domElement.removeEventListener('mousedown', onMouseDown, false);
              domElement.removeEventListener('wheel', onMouseWheel, false);
              domElement.removeEventListener('mousemove', onMouseMove, false);
              domElement.removeEventListener('mouseup', onMouseUp, false);
              domElement.removeEventListener('mouseout', onMouseUp, false);
              domElement.removeEventListener('dblclick', onMouseUp, false);
              domElement.removeEventListener('touchstart', touchStart, false);
              domElement.removeEventListener('touchmove', touchMove, false);
          };
      };
      MapControls2D.prototype = Object.create(EventDispatcher.prototype);
      MapControls2D.prototype.constructor = MapControls2D;

      const ResizeObserver = window.ResizeObserver || ResizeObserver$2;

      class Viewer2d {
        constructor(scene, options = {}) {
          this.options = {
            ...DEFAULT_OPTIONS_SCENE,
            ...options,
          };
          this.scene = scene || new Scene();
          this.previousWidth = 0;
          this.previousHeight = 0;

          this.ro = new ResizeObserver(entries => {
            if (entries.length > 0) {
              this.onWindowResize();
            }
          });

          this.initCamera();
          this.animate = this.animate.bind(this);
          this.onWindowResize = this.onWindowResize.bind(this);
        }

        animate() {
          this.scene.stats.begin();
          this.scene.updateMatrixWorld();

          this.renderer.render(this.scene, this.camera);
          this.scene.stats.end();
          requestAnimationFrame(this.animate);
        }

        initCamera() {
          this.camera = new OrthographicCamera(-100, 100, 100, -100, 0.1, 1000);
          this.camera.zoom = 0.5;
          this.camera.position.set(0, 0, 10);
          this.camera.up = new Vector3(0, 0, 1);
          this.camera.lookAt(new Vector3());

          this.scene.add(this.camera);
        }

        setContainer(domNode) {
          this.container = domNode;
          this.initRenderer(domNode);
          this.controls = new MapControls2D(this.camera, this.container);
          this.controls.enableDamping = true;
          this.ro.observe(this.container);
          requestAnimationFrame(this.animate);
          this.onWindowResize();
        }

        initRenderer() {
          const renderer = new WebGLRenderer({ antialias: true });
          renderer.autoClear = false;
          renderer.autoUpdateScene = false;
          renderer.setPixelRatio(window.devicePixelRatio);
          renderer.setSize(this.container.offsetWidth, this.container.offsetHeight);
          this.renderer = renderer;
          this.container.appendChild(renderer.domElement);
        }

        updateOptions(options) {
          this.options = {
            ...this.options,
            ...options,
          };
          this.scene.updateOptions(this.options);
        }

        destroy() {
          this.ro.unobserve(this.container);
        }

        onWindowResize() {
          const { camera } = this;
          const { offsetHeight, offsetWidth } = this.container;
          if (
            Math.abs(offsetWidth - this.previousWidth) > 10 ||
            Math.abs(offsetHeight - this.previousHeight) > 10
          ) {
            const [cameraWidth, cameraHeight] = [
              offsetWidth / 100,
              offsetHeight / 100,
            ];
            camera.left = -cameraWidth / 2;
            camera.right = cameraWidth / 2;
            camera.top = cameraHeight / 2;
            camera.bottom = -cameraHeight / 2;
            camera.updateProjectionMatrix();
            this.renderer.setSize(offsetWidth, offsetHeight);
            this.previousWidth = offsetWidth;
            this.previousHeight = offsetHeight;
          }
        }

        addVisualization(vizObject) {
          this.scene.addVisualization(vizObject);
        }
      }

      // @ts-nocheck
      var EditorControls = function (object, domElement) {
          domElement = domElement !== undefined ? domElement : document;
          // API
          this.enabled = true;
          this.center = new Vector3();
          this.panSpeed = 0.002;
          this.zoomSpeed = 0.1;
          this.rotationSpeed = 0.005;
          // internals
          var scope = this;
          var vector = new Vector3();
          var delta = new Vector3();
          var box = new Box3();
          var STATE = { NONE: -1, ROTATE: 0, ZOOM: 1, PAN: 2 };
          var state = STATE.NONE;
          var center = this.center;
          var normalMatrix = new Matrix3();
          var pointer = new Vector2();
          var pointerOld = new Vector2();
          var spherical = new Spherical();
          var sphere = new Sphere$1();
          // events
          var changeEvent = { type: 'change' };
          this.focus = function (target) {
              var distance;
              box.setFromObject(target);
              if (box.isEmpty() === false) {
                  box.getCenter(center);
                  distance = box.getBoundingSphere(sphere).radius;
              }
              else {
                  // Focusing on an Group, AmbientLight, etc
                  center.setFromMatrixPosition(target.matrixWorld);
                  distance = 0.1;
              }
              delta.set(0, 0, 1);
              delta.applyQuaternion(object.quaternion);
              delta.multiplyScalar(distance * 4);
              object.position.copy(center).add(delta);
              scope.dispatchEvent(changeEvent);
          };
          this.pan = function (delta) {
              var distance = object.position.distanceTo(center);
              delta.multiplyScalar(distance * scope.panSpeed);
              delta.applyMatrix3(normalMatrix.getNormalMatrix(object.matrix));
              object.position.add(delta);
              center.add(delta);
              scope.dispatchEvent(changeEvent);
          };
          this.zoom = function (delta) {
              var distance = object.position.distanceTo(center);
              delta.multiplyScalar(distance * scope.zoomSpeed);
              if (delta.length() > distance)
                  return;
              delta.applyMatrix3(normalMatrix.getNormalMatrix(object.matrix));
              object.position.add(delta);
              scope.dispatchEvent(changeEvent);
          };
          this.rotate = function (delta) {
              vector.copy(object.position).sub(center);
              // spherical.setFromVector3( vector );
              spherical.setFromCartesianCoords(-1 * vector.x, vector.z, vector.y);
              spherical.theta += delta.x * scope.rotationSpeed;
              spherical.phi += delta.y * scope.rotationSpeed;
              spherical.makeSafe();
              var tempRelPosition = vector.setFromSpherical(spherical);
              vector.set(-1 * tempRelPosition.x, tempRelPosition.z, tempRelPosition.y);
              object.position.copy(center).add(vector);
              object.lookAt(center);
              scope.dispatchEvent(changeEvent);
          };
          // mouse
          function onMouseDown(event) {
              if (scope.enabled === false)
                  return;
              if (event.button === 0) {
                  state = STATE.ROTATE;
              }
              else if (event.button === 1) {
                  state = STATE.ZOOM;
              }
              else if (event.button === 2) {
                  state = STATE.PAN;
              }
              pointerOld.set(event.clientX, event.clientY);
              domElement.addEventListener('mousemove', onMouseMove, false);
              domElement.addEventListener('mouseup', onMouseUp, false);
              domElement.addEventListener('mouseout', onMouseUp, false);
              domElement.addEventListener('dblclick', onMouseUp, false);
          }
          function onMouseMove(event) {
              if (scope.enabled === false)
                  return;
              pointer.set(event.clientX, event.clientY);
              var movementX = pointer.x - pointerOld.x;
              var movementY = pointer.y - pointerOld.y;
              if (state === STATE.ROTATE) {
                  scope.rotate(delta.set(-movementX, -movementY, 0));
              }
              else if (state === STATE.ZOOM) {
                  scope.zoom(delta.set(0, 0, movementY));
              }
              else if (state === STATE.PAN) {
                  scope.pan(delta.set(-movementX, movementY, 0));
              }
              pointerOld.set(event.clientX, event.clientY);
          }
          function onMouseUp(event) {
              domElement.removeEventListener('mousemove', onMouseMove, false);
              domElement.removeEventListener('mouseup', onMouseUp, false);
              domElement.removeEventListener('mouseout', onMouseUp, false);
              domElement.removeEventListener('dblclick', onMouseUp, false);
              state = STATE.NONE;
          }
          function onMouseWheel(event) {
              event.preventDefault();
              // Normalize deltaY due to https://bugzilla.mozilla.org/show_bug.cgi?id=1392460
              scope.zoom(delta.set(0, 0, event.deltaY > 0 ? 1 : -1));
          }
          function contextmenu(event) {
              event.preventDefault();
          }
          this.dispose = function () {
              domElement.removeEventListener('contextmenu', contextmenu, false);
              domElement.removeEventListener('mousedown', onMouseDown, false);
              domElement.removeEventListener('wheel', onMouseWheel, false);
              domElement.removeEventListener('mousemove', onMouseMove, false);
              domElement.removeEventListener('mouseup', onMouseUp, false);
              domElement.removeEventListener('mouseout', onMouseUp, false);
              domElement.removeEventListener('dblclick', onMouseUp, false);
              domElement.removeEventListener('touchstart', touchStart, false);
              domElement.removeEventListener('touchmove', touchMove, false);
          };
          domElement.addEventListener('contextmenu', contextmenu, false);
          domElement.addEventListener('mousedown', onMouseDown, false);
          domElement.addEventListener('wheel', onMouseWheel, false);
          // touch
          var touches = [new Vector3(), new Vector3(), new Vector3()];
          var prevTouches = [new Vector3(), new Vector3(), new Vector3()];
          var prevDistance = null;
          function touchStart(event) {
              if (scope.enabled === false)
                  return;
              switch (event.touches.length) {
                  case 1:
                      touches[0]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      touches[1]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      break;
                  case 2:
                      touches[0]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      touches[1]
                          .set(event.touches[1].pageX, event.touches[1].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      prevDistance = touches[0].distanceTo(touches[1]);
                      break;
              }
              prevTouches[0].copy(touches[0]);
              prevTouches[1].copy(touches[1]);
          }
          function touchMove(event) {
              if (scope.enabled === false)
                  return;
              event.preventDefault();
              event.stopPropagation();
              function getClosest(touch, touches) {
                  var closest = touches[0];
                  for (var i in touches) {
                      if (closest.distanceTo(touch) > touches[i].distanceTo(touch)) {
                          closest = touches[i];
                      }
                  }
                  return closest;
              }
              switch (event.touches.length) {
                  case 1:
                      touches[0]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      touches[1]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      scope.rotate(touches[0]
                          .sub(getClosest(touches[0], prevTouches))
                          .multiplyScalar(-1));
                      break;
                  case 2:
                      touches[0]
                          .set(event.touches[0].pageX, event.touches[0].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      touches[1]
                          .set(event.touches[1].pageX, event.touches[1].pageY, 0)
                          .divideScalar(window.devicePixelRatio);
                      var distance = touches[0].distanceTo(touches[1]);
                      scope.zoom(delta.set(0, 0, prevDistance - distance));
                      prevDistance = distance;
                      var offset0 = touches[0]
                          .clone()
                          .sub(getClosest(touches[0], prevTouches));
                      var offset1 = touches[1]
                          .clone()
                          .sub(getClosest(touches[1], prevTouches));
                      offset0.x = -offset0.x;
                      offset1.x = -offset1.x;
                      scope.pan(offset0.add(offset1));
                      break;
              }
              prevTouches[0].copy(touches[0]);
              prevTouches[1].copy(touches[1]);
          }
          domElement.addEventListener('touchstart', touchStart, false);
          domElement.addEventListener('touchmove', touchMove, false);
      };
      EditorControls.prototype = Object.create(EventDispatcher.prototype);
      EditorControls.prototype.constructor = EditorControls;

      const ResizeObserver$1 = window.ResizeObserver || ResizeObserver$2;

      class Viewer3d {
        constructor(scene, options = {}) {
          this.options = {
            ...DEFAULT_OPTIONS_SCENE,
            ...options,
          };
          this.scene = scene || new Scene(this.options);
          this.previousWidth = 0;
          this.previousHeight = 0;

          this.ro = new ResizeObserver$1(entries => {
            if (entries.length > 0) {
              this.onWindowResize();
            }
          });

          this.initCamera();
          this.animate = this.animate.bind(this);
          this.onWindowResize = this.onWindowResize.bind(this);
        }

        animate() {
          this.scene.stats.begin();
          this.scene.updateMatrixWorld();

          this.renderer.render(this.scene, this.camera);
          this.scene.stats.end();
          requestAnimationFrame(this.animate);
        }

        initCamera() {
          this.camera = new PerspectiveCamera(50, 1, 0.01, 1000);
          this.camera.position.set(0, 5, 10);
          this.camera.up = new Vector3(0, 0, 1);
          this.camera.lookAt(new Vector3());

          this.scene.add(this.camera);
        }

        setContainer(domNode) {
          this.container = domNode;
          this.initRenderer(domNode);
          this.controls = new EditorControls(this.camera, this.container);
          this.controls.enableDamping = true;
          this.ro.observe(this.container);
          requestAnimationFrame(this.animate);
          this.onWindowResize();
        }

        initRenderer() {
          const renderer = new WebGLRenderer({ antialias: true });
          renderer.autoClear = false;
          renderer.autoUpdateScene = false;
          renderer.setPixelRatio(window.devicePixelRatio);
          renderer.setSize(this.container.offsetWidth, this.container.offsetHeight);
          this.renderer = renderer;
          this.container.appendChild(renderer.domElement);
        }

        updateOptions(options) {
          this.options = {
            ...this.options,
            ...options,
          };
          this.scene.updateOptions(this.options);
        }

        destroy() {
          this.ro.unobserve(this.container);
        }

        onWindowResize() {
          const { camera } = this;
          const { offsetHeight, offsetWidth } = this.container;
          if (
            Math.abs(offsetWidth - this.previousWidth) > 10 ||
            Math.abs(offsetHeight - this.previousHeight) > 10
          ) {
            camera.aspect = offsetWidth / offsetHeight;
            camera.updateProjectionMatrix();
            this.renderer.setSize(offsetWidth, offsetHeight);
            this.previousWidth = offsetWidth;
            this.previousHeight = offsetHeight;
          }
        }

        addVisualization(vizObject) {
          this.scene.addVisualization(vizObject);
        }
      }

      class TfViewer extends Viewer3d {
        constructor(rosInstance, options) {
          super(null, {
            ...DEFAULT_OPTIONS_TF_VIEWER,
            ...options,
          });
          const { onFramesListUpdate } = this.options;
          this.ros = rosInstance;
          this.framesList = [];
          this.onFramesListUpdate = onFramesListUpdate || (() => {});

          this.initRosEvents();
          this.getTFMessages = this.getTFMessages.bind(this);
          this.setFrameTransform = this.setFrameTransform.bind(this);
          this.addRobot = this.addRobot.bind(this);
          this.onRosConnection = this.onRosConnection.bind(this);
        }

        initRosEvents() {
          this.ros.on('connection', () => {
            this.onRosConnection();
          });
          if (this.ros.isConnected) {
            this.onRosConnection();
          }
        }

        onRosConnection() {
          this.ros.getTopics(rosTopics => {
            ['/tf', '/tf_static'].forEach(name => {
              const topic = new ROSLIB.Topic({
                ros: this.ros,
                name,
                messageType: rosTopics.types[rosTopics.topics.indexOf(name)],
              });
              topic.subscribe(this.getTFMessages);
            });
          });
        }

        getTFMessages({ transforms }) {
          transforms.forEach(
            ({
              header: { frame_id: parentFrameId },
              child_frame_id: childFrameId,
              transform: {
                rotation: { w: rw, x: rx, y: ry, z: rz },
                translation: { x, y, z },
              },
            }) => {
              const [childObject, parentObject] = [
                this.getObjectOrCreate(childFrameId),
                this.getObjectOrCreate(parentFrameId),
              ];

              parentObject.add(childObject);
              childObject.position.set(x, y, z);
              childObject.quaternion.set(rx, ry, rz, rw);

              [parentFrameId, childFrameId].forEach(frame => {
                if (this.framesList.indexOf(frame) === -1) {
                  this.framesList.push(frame);
                }
              });
            },
          );
          this.setFrameTransform();
        }

        getObjectOrCreate(frameId) {
          if (this.framesList.indexOf(frameId) === -1) {
            this.framesList.push(frameId);
            this.onFramesListUpdate(this.framesList);
          }
          const existingFrame = this.scene.getObjectByName(frameId);
          if (existingFrame) {
            return existingFrame;
          }

          const newFrame = new Group$1();
          newFrame.name = frameId;
          this.scene.addObject(newFrame);
          return newFrame;
        }

        setFrameTransform() {
          const {
            options: { selectedFrame },
            scene: { vizWrapper },
          } = this;
          if (!selectedFrame) {
            return;
          }
          const currentFrameObject = vizWrapper.getObjectByName(selectedFrame);

          if (currentFrameObject) {
            vizWrapper.position.set(0, 0, 0);
            vizWrapper.quaternion.set(0, 0, 0, 1);
            vizWrapper.updateMatrixWorld();

            const worldPos = new Vector3();
            const worldQuat = new Quaternion();

            currentFrameObject.getWorldQuaternion(worldQuat);
            const { w: quatw, x: quatx, y: quaty, z: quatz } = worldQuat;
            vizWrapper.quaternion.set(-quatx, -quaty, -quatz, quatw);

            vizWrapper.updateMatrixWorld();

            currentFrameObject.getWorldPosition(worldPos);
            const oppPos = worldPos.negate();
            vizWrapper.position.set(oppPos.x, oppPos.y, oppPos.z);
          }
        }

        addVisualization(vizObject) {
          super.addVisualization(vizObject);

          vizObject.onHeaderChange = newFrameId => {
            const frameObject = this.getObjectOrCreate(newFrameId);
            frameObject.add(vizObject.object);
          };
        }

        attachObjectOutsideTree(object) {
          const frameObject = this.getObjectOrCreate(object.frameId);
          frameObject.attach(object);
        }

        addRobot(robotModel) {
          robotModel.load(object => {
            RobotModel.onComplete(object);
            super.addVisualization(robotModel);
            // eslint-disable-next-line guard-for-in
            for (const linkName in object.children[0].links) {
              const o = object.children[0].links[linkName];
              const existingObject = this.scene.getObjectByName(o.name);
              if (existingObject) {
                o.children.forEach(child => {
                  existingObject.add(child);
                });
              }
            }
          });
        }

        setFixedFrame() {}
      }

      class InteractiveMarkerManager {
        constructor(rootObject, viewer, options) {
          this.markerManagerMap = {};
          this.contolsManagerMap = {};
          this.objectMap = {};
          this.object = rootObject;
          this.viewer = viewer;
          this.hideOtherHandlesOnSelect = options.hideOtherHandlesOnSelect;
          this.hideOtherControlsInstancesOnSelect =
            options.hideOtherControlsInstancesOnSelect;
          this.visible = options.visible;
          this.namespaces = {};
          this.onChange = this.onChange.bind(this);
        }

        getMarkerManagerOrCreate(interactiveMarkerName) {
          const id = interactiveMarkerName;
          if (!this.markerManagerMap[id] || !this.objectMap[id]) {
            const markersHolder = new Group();
            markersHolder.name = id || '';
            this.markerManagerMap[id] = new MarkerManager(
              markersHolder,
              this.onChange,
            );
            this.objectMap[id] = markersHolder;
            this.object.add(markersHolder);
          }

          return {
            manager: this.markerManagerMap[id],
            object: this.objectMap[id],
          };
        }

        onChange() {
          if (this.callback) {
            this.callback();
          }
        }

        onNamespaceChange(callback) {
          this.callback = callback;
        }

        initMarkers(interactiveMarker, freeformControls, visible) {
          const {
            controls,
            header: { frame_id },
            name,
            pose: { orientation, position },
            scale,
          } = interactiveMarker;
          const { manager, object } = this.getMarkerManagerOrCreate(name);

          manager.visible = visible;
          object.visible = visible;

          object.frameId = frame_id;
          if (this.viewer.attachObjectOutsideTree) {
            this.viewer.attachObjectOutsideTree(object);
          }

          object.setTransform({ translation: position, rotation: orientation });
          object.userData.control = {
            frameId: interactiveMarker.header.frame_id,
            markerName: interactiveMarker.name,
          };
          object.userData.handlesControlsMap = {};

          controls.forEach((control, index) => {
            // cannot rely on the control.name being present or unique
            const key = name;
            if (this.contolsManagerMap[key] === undefined) {
              this.contolsManagerMap[key] = [];
            }
            if (this.contolsManagerMap[key].length !== controls.length) {
              const attachMode =
                control.orientation_mode ===
                INTERACTIVE_MARKER_ORIENTATION_MODES.FIXED
                  ? ANCHOR_MODE.FIXED
                  : ANCHOR_MODE.INHERIT;
              const controlsManager = freeformControls.anchor(object, {
                separationT: {
                  x: 0.4,
                  y: 0.4,
                  z: 0.4,
                },
                orientation: control.orientation,
                mode: attachMode,
                hideOtherHandlesOnSelect: this.hideOtherHandlesOnSelect,
                hideOtherControlsInstancesOnSelect: this
                  .hideOtherControlsInstancesOnSelect,
                showHelperPlane: true,
              });
              controlsManager.scale.set(scale, scale, scale);
              controlsManager.visible = visible;

              InteractiveMarkerManager.enableControls(
                object,
                control.interaction_mode,
                control.orientation_mode,
                control.orientation,
                control.name,
                controlsManager,
                randomColor({
                  seed: `${key}-${control.name}-${2 * index}`, // 2 * for more variable color
                }),
              );

              this.contolsManagerMap[key].push(controlsManager);
            }

            if (control.markers.length > 0) {
              control.markers.forEach(marker => {
                manager.updateMarker(marker);
              });
            }
          });
        }

        setVisible(visible) {
          this.visible = visible;
          const controlMangerGroups = Object.values(this.contolsManagerMap);
          controlMangerGroups.map(managers => {
            managers.map(manager => {
              manager.visible = visible;
              manager.object.visible = visible;
            });
          });
        }

        updatePose(poseObject) {
          const { name, pose } = poseObject;
          const markerObject = this.objectMap[name];
          if (markerObject && pose) {
            const { orientation, position } = pose;
            markerObject.setTransform({
              translation: position,
              rotation: orientation,
            });
          }
        }

        static enableControls(
          object,
          interactionMode,
          orientationMode,
          orientation,
          controlName,
          controlsManager,
          givenColor,
        ) {
          controlsManager.showAll(false);

          const controlsManagerOrientation = new Quaternion();
          controlsManager.getWorldQuaternion(controlsManagerOrientation);
          const controlsManagerRotation = new Euler().setFromQuaternion(
            controlsManagerOrientation,
          );
          const alignmentColor = InteractiveMarkerManager.getAlignmentColor(
            controlsManagerRotation,
          );
          const color = alignmentColor || givenColor;

          // currently only ROTATE_AXIS is supported for VIEW_FACING mode
          // this is the most useful one
          if (
            UNSUPPORTED_INTERACTIVE_MARKER_ORIENTATION_MODES.indexOf(
              orientationMode,
            ) !== -1 &&
            interactionMode === INTERACTIVE_MARKER_INTERACTION_MODES.ROTATE_AXIS
          ) {
            object.userData.handlesControlsMap = {
              ...object.userData.handlesControlsMap,
              [DEFAULT_HANDLE_GROUP_NAME.ER]: controlName,
            };
            controlsManager.showByNames([DEFAULT_HANDLE_GROUP_NAME.ER], true);
            return;
          }

          let handles = [];

          switch (interactionMode) {
            case INTERACTIVE_MARKER_INTERACTION_MODES.NONE:
            case INTERACTIVE_MARKER_INTERACTION_MODES.MENU:
            case INTERACTIVE_MARKER_INTERACTION_MODES.BUTTON: {
              break;
            }
            case INTERACTIVE_MARKER_INTERACTION_MODES.MOVE_AXIS: {
              handles = [
                DEFAULT_HANDLE_GROUP_NAME.XPT,
                DEFAULT_HANDLE_GROUP_NAME.XNT,
              ];

              if (color !== undefined) {
                controlsManager.translationXP.setColor(color);
                controlsManager.translationXN.setColor(color);
              }

              break;
            }
            case INTERACTIVE_MARKER_INTERACTION_MODES.MOVE_PLANE: {
              handles = [DEFAULT_HANDLE_GROUP_NAME.PICK_PLANE_YZ];

              if (color !== undefined) {
                controlsManager.pickPlaneYZ.setColor(color);
              }

              break;
            }
            case INTERACTIVE_MARKER_INTERACTION_MODES.ROTATE_AXIS: {
              handles = [DEFAULT_HANDLE_GROUP_NAME.XR];

              if (color !== undefined) {
                controlsManager.rotationX.setColor(color);
              }

              break;
            }
            case INTERACTIVE_MARKER_INTERACTION_MODES.MOVE_ROTATE: {
              handles = [
                DEFAULT_HANDLE_GROUP_NAME.XR,
                DEFAULT_HANDLE_GROUP_NAME.PICK_PLANE_YZ,
              ];

              if (color !== undefined) {
                controlsManager.pickPlaneYZ.setColor(color);
                controlsManager.rotationX.setColor(color);
              }

              break;
            }
            case INTERACTIVE_MARKER_INTERACTION_MODES.MOVE_3D: {
              handles = [
                DEFAULT_HANDLE_GROUP_NAME.XPT,
                DEFAULT_HANDLE_GROUP_NAME.XNT,
                DEFAULT_HANDLE_GROUP_NAME.YPT,
                DEFAULT_HANDLE_GROUP_NAME.YNT,
                DEFAULT_HANDLE_GROUP_NAME.ZPT,
                DEFAULT_HANDLE_GROUP_NAME.ZNT,
                DEFAULT_HANDLE_GROUP_NAME.PICK,
              ];
              break;
            }
            case INTERACTIVE_MARKER_INTERACTION_MODES.ROTATE_3D: {
              handles = [
                DEFAULT_HANDLE_GROUP_NAME.XR,
                DEFAULT_HANDLE_GROUP_NAME.YR,
                DEFAULT_HANDLE_GROUP_NAME.ZR,
              ];
              break;
            }
            case INTERACTIVE_MARKER_INTERACTION_MODES.MOVE_ROTATE_3D: {
              // these handles are currently commented out
              // because these overlap with the extra 6-DOF controls
              // that are available in the example and also in kompose
              // proper fix requires increasing the scale of the below
              // handles so that they don't overlap
              handles = [
                // DEFAULT_HANDLE_GROUP_NAME.XPT,
                // DEFAULT_HANDLE_GROUP_NAME.XNT,
                // DEFAULT_HANDLE_GROUP_NAME.YPT,
                // DEFAULT_HANDLE_GROUP_NAME.YNT,
                // DEFAULT_HANDLE_GROUP_NAME.ZPT,
                // DEFAULT_HANDLE_GROUP_NAME.ZNT,
                // DEFAULT_HANDLE_GROUP_NAME.XR,
                // DEFAULT_HANDLE_GROUP_NAME.YR,
                // DEFAULT_HANDLE_GROUP_NAME.ZR,
                DEFAULT_HANDLE_GROUP_NAME.PICK,
              ];
              break;
            }
          }

          handles.map(handle => {
            object.userData.handlesControlsMap[handle] = controlName;
          });
          controlsManager.showByNames(handles, true);
        }

        static getAlignmentColor(rotation) {
          const threshold = 10 ** -3;
          const [x, y, z] = rotation.toArray();
          for (let i = -4; i <= 4; i++) {
            const angle = (i * Math.PI) / 2;
            if (
              Math.abs(x - angle) < threshold &&
              Math.abs(y) < threshold &&
              Math.abs(z) < threshold
            ) {
              return 'red';
            }
            if (
              Math.abs(y - angle) < threshold &&
              Math.abs(x) < threshold &&
              Math.abs(z) < threshold
            ) {
              return 'green';
            }
            if (
              Math.abs(z - angle) < threshold &&
              Math.abs(y) < threshold &&
              Math.abs(x) < threshold
            ) {
              return 'blue';
            }
          }

          return null;
        }

        removeObject(id) {
          const obj = this.objectMap[id];
          obj.parent.remove(obj);
          delete this.objectMap[id];
        }

        reset(destroy) {
          this.onChange();

          if (destroy) {
            Object.keys(this.objectMap).forEach(id => {
              this.removeObject(id);
            });
          }
        }
      }

      class InteractiveMarkers extends Core {
        constructor(
          ros,
          topicName,
          viewer,
          options = DEFAULT_OPTIONS_INTERACTIVE_MARKER,
        ) {
          super(ros, topicName, MESSAGE_TYPE_INTERACTIVEMARKER, {
            ...DEFAULT_OPTIONS_INTERACTIVE_MARKER,
            ...options,
          });

          this.init = false;
          this.object = new Group();
          this.viewer = viewer;

          this.updateOptions({
            ...DEFAULT_OPTIONS_INTERACTIVE_MARKER,
            ...options,
          });

          this.interactiveMarkerManager = new InteractiveMarkerManager(
            this.object,
            viewer,
            this.options,
          );

          this.interactiveMarkersNames = new Set();
          this.interactiveMarkersFrameIds = new Set();
          this.clientId = `amphion-${Math.round(Math.random() * 10 ** 8)}`;
          this.messageSequence = 0;
          this.feedbackTopic = null;

          const { publishThrottleRate } = this.options;
          this.debouncedPublish = debounce(
            this.publish.bind(this),
            publishThrottleRate && publishThrottleRate > 0 ? publishThrottleRate : 0,
          );
          this.initFreeformControls();
          this.publishManual = this.publishManual.bind(this);
        }

        hide() {
          super.hide();
          this.interactiveMarkerManager.setVisible(false);
        }

        show() {
          super.show();
          this.interactiveMarkerManager.setVisible(true);
        }

        destroy() {
          super.destroy();
          this.freeformControls.destroy();
          this.freeformControls = null;
          this.interactiveMarkerManager.reset(true);
          this.interactiveMarkersNames = new Set();
        }

        initFreeformControls() {
          const { camera, controls, renderer, scene } = this.viewer;
          this.freeformControls = new ControlsManager(camera, renderer.domElement);
          scene.add(this.freeformControls);

          this.freeformControls.listen(RAYCASTER_EVENTS.DRAG_START, () => {
            controls.enabled = false;
          });

          this.freeformControls.listen(RAYCASTER_EVENTS.DRAG, (object, handleName) =>
            this.debouncedPublish(object, handleName),
          );

          this.freeformControls.listen(
            RAYCASTER_EVENTS.DRAG_STOP,
            (object, handleName) => {
              this.debouncedPublish(object, handleName);
              controls.enabled = true;
            },
          );
        }

        static makeInteractiveMarkerFeedbackMessage({
          seq,
          frame_id: frameId,
          client_id: clientId,
          marker_name: markerName,
          control_name: controlName,
          position,
          quaternion,
        }) {
          return new ROSLIB.Message({
            header: {
              seq,
              frame_id: frameId,
              stamp: {
                secs: 0,
                nsecs: 0,
              },
            },
            client_id: clientId,
            marker_name: markerName,
            control_name: controlName,
            pose: {
              position: {
                x: position.x,
                y: position.y,
                z: position.z,
              },
              orientation: {
                x: quaternion.x,
                y: quaternion.y,
                z: quaternion.z,
                w: quaternion.w,
              },
            },
            event_type: 1,
            menu_entry_id: 0,
            mouse_point: {
              x: 0,
              y: 0,
              z: 0,
            },
            mouse_point_valid: false,
          });
        }

        static makeInteractiveMarkerFeedbackTopic(ros, name) {
          return new ROSLIB.Topic({
            ros,
            name,
            messageType: MESSAGE_TYPE_INTERACTIVEMARKER_FEEDBACK,
          });
        }

        publish(object, handleName) {
          if (!object) {
            return;
          }

          const { frameId, markerName } = object.userData.control;
          const controlName = object.userData.handlesControlsMap[handleName];

          const message = InteractiveMarkers.makeInteractiveMarkerFeedbackMessage({
            seq: this.messageSequence,
            client_id: this.clientId,
            frame_id: frameId,
            marker_name: markerName,
            control_name: controlName,
            position: object.position,
            quaternion: object.quaternion,
          });

          if (this.feedbackTopic !== null) {
            this.feedbackTopic.publish(message);
          }

          this.messageSequence++;
        }

        publishManual(pose) {
          const message = InteractiveMarkers.makeInteractiveMarkerFeedbackMessage({
            seq: this.messageSequence,
            client_id: this.clientId,
            ...pose,
            frame_id: Array.from(this.interactiveMarkersFrameIds)[0],
            marker_name: Array.from(this.interactiveMarkersNames)[0],
          });

          if (this.feedbackTopic !== null) {
            this.feedbackTopic.publish(message);
          }

          this.messageSequence++;
        }

        updateOptions(options) {
          if (options.feedbackTopicName !== undefined) {
            if (
              !this.feedbackTopic ||
              this.feedbackTopic.name !== options.feedbackTopicName.name
            ) {
              this.feedbackTopic = InteractiveMarkers.makeInteractiveMarkerFeedbackTopic(
                this.ros,
                options.feedbackTopicName.name,
              );
            }
          } else {
            this.feedbackTopic = null;
          }

          // need a better way to handle interdependent topics
          const shouldSubscriptionChange =
            this.options.updateTopicName !== options.topicName && this.init;
          const guardAgainstOtherOptionsChange =
            this.topicName === this.options.updateTopicName;

          if (shouldSubscriptionChange && options.updateTopicName !== undefined) {
            const { messageType, name } = options.updateTopicName;
            this.changeTopic(name, messageType, true);
          } else if (shouldSubscriptionChange && guardAgainstOtherOptionsChange) {
            this.unsubscribe();
          }
          super.updateOptions(options);
        }

        update(message) {
          super.update(message);
          if (message.markers && message.markers.length > 0) {
            message.markers.forEach(interactiveMarker => {
              if (!this.interactiveMarkersNames.has(interactiveMarker.name)) {
                this.interactiveMarkerManager.initMarkers(
                  interactiveMarker,
                  this.freeformControls,
                  this.options.visible,
                );
              }
              this.interactiveMarkersNames.add(interactiveMarker.name);
              if (interactiveMarker.header) {
                this.interactiveMarkersFrameIds.add(
                  interactiveMarker.header.frame_id,
                );
              }

              message.markers.forEach(poseObject => {
                this.interactiveMarkerManager.updatePose(poseObject);
              });
            });
            // need a better way to handle interdependent topics
            if (!this.init) {
              this.init = true;
              if (this.options.updateTopicName !== undefined) {
                const { messageType, name } = this.options.updateTopicName;
                this.init = true;
                this.changeTopic(name, messageType, true);
              } else {
                this.unsubscribe();
              }
            }
          }

          // for InteractiveMarkerUpdate sub-message (InteractiveMarkerPose)
          if (message.poses && message.poses.length > 0) {
            message.poses.forEach(pose => {
              this.interactiveMarkerManager.updatePose(pose);
            });
          }

          // for InteractiveMarkerUpdate sub-message
          if (message.erases && message.poses.leading > 0) ;
        }

        reset() {
          this.interactiveMarkerManager.reset(false);
        }
      }

      var index = exports('default', {
          CollisionObject: CollisionObject,
          DepthCloud: DepthCloud,
          DisplayTrajectory: DisplayTrajectory,
          PlanningScene: PlanningScene,
          Point: Point,
          PointCloud: PointCloud,
          Polygon: Polygon,
          Pose: Pose,
          Wrench: Wrench,
          PoseArray: PoseArray,
          Tf: Tf,
          RobotModel: RobotModel,
          MarkerArray: MarkerArray,
          LaserScan: LaserScan,
          Map: Map$1,
          Odometry: Odometry,
          Path: Path,
          Image: Image,
          ImageStream: ImageStream,
          Marker: Marker,
          Range: Range,
          Scene: Scene,
          Viewer2d: Viewer2d,
          Viewer3d: Viewer3d,
          TfViewer: TfViewer,
          InteractiveMarkers: InteractiveMarkers,
          CONSTANTS: CONSTANTS,
      });

    }
  };
});
//# sourceMappingURL=index.js.map
